# Licensed under a 3-clause BSD style license - see LICENSE.rst
# This module implements the base processing for HRS data

# See

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
import numpy as np
from astropy.extern import six

from astropy import units as u
import scipy.ndimage as nd

import ccdproc

from .hrsorder import HRSOrder

__all__ = ['ccd_process', 'create_masterbias', 'hrs_process', 'blue_process',
           'red_process', 'flatfield_science_order', 'flatfield_science']


def ccd_process(ccd, oscan=None, trim=None, error=False, masterbias=None,
                bad_pixel_mask=None, gain=None, rdnoise=None,
                oscan_median=True, oscan_model=None):
    """Perform basic processing on ccd data.

       The following steps can be included:
        * overscan correction
        * trimming of the image
        * create edeviation frame
        * gain correction
        * add a mask to the data
        * subtraction of master bias

       The task returns a processed `ccdproc.CCDData` object.

    Parameters
    ----------
    ccd: `ccdproc.CCDData`
        Frame to be reduced

    oscan: None, str, or, `~ccdproc.ccddata.CCDData`
        For no overscan correction, set to None.   Otherwise proivde a region
        of `ccd` from which the overscan is extracted, using the FITS
        conventions for index order and index start, or a
        slice from `ccd` that contains the overscan.

    trim: None or str
        For no trim correction, set to None.   Otherwise proivde a region
        of `ccd` from which the image should be trimmed, using the FITS
        conventions for index order and index start.

    error: boolean
        If True, create an uncertainty array for ccd

    masterbias: None, `~numpy.ndarray`,  or `~ccdproc.CCDData`
        A materbias frame to be subtracted from ccd.

    bad_pixel_mask: None or `~numpy.ndarray`
        A bad pixel mask for the data. The bad pixel mask should be in given
        such that bad pixels havea value of 1 and good pixels a value of 0.

    gain: None or `~astropy.Quantity`
        Gain value to multiple the image by to convert to electrons

    rdnoise: None or `~astropy.Quantity`
        Read noise for the observations.  The read noise should be in
        `~astropy.units.electron`


    oscan_median :  bool, optional
        If true, takes the median of each line.  Otherwise, uses the mean

    oscan_model :  `~astropy.modeling.Model`, optional
        Model to fit to the data.  If None, returns the values calculated
        by the median or the mean.

    Returns
    -------
    ccd: `ccdproc.CCDData`
        Reduded ccd

    Examples
    --------

    1. To overscan, trim, and gain correct a data set:

    >>> import numpy as np
    >>> from astropy import units as u
    >>> from hrsprocess import ccd_process
    >>> ccd = CCDData(np.ones([100, 100]), unit=u.adu)
    >>> nccd = ccd_process(ccd, oscan='[1:10,1:100]', trim='[10:100, 1,100]',
                           error=False, gain=2.0*u.electron/u.adu)


    """
    # make a copy of the object
    nccd = ccd.copy()

    # apply the overscan correction
    if isinstance(oscan, ccdproc.CCDData):
        nccd = ccdproc.subtract_overscan(nccd, overscan=oscan,
                                         median=oscan_median,
                                         model=oscan_model)
    elif isinstance(oscan, six.string_types):
        nccd = ccdproc.subtract_overscan(nccd, fits_section=oscan,
                                         median=oscan_median,
                                         model=oscan_model)
    elif oscan is None:
        pass
    else:
        raise TypeError('oscan is not None, a string, or CCDData object')

    # apply the trim correction
    if isinstance(trim, six.string_types):
        nccd = ccdproc.trim_image(nccd, fits_section=trim)
    elif trim is None:
        pass
    else:
        raise TypeError('trim is not None or a string')

    # test subtracting the master bias
    if isinstance(masterbias, ccdproc.CCDData):
        nccd = ccdproc.subtract_bias(nccd, masterbias)
    elif isinstance(masterbias, np.ndarray):
        nccd.data = nccd.data - masterbias
    elif masterbias is None:
        pass
    else:
        raise TypeError(
            'masterbias is not None, numpy.ndarray,  or a CCDData object')

    # create the error frame
    if error and gain is not None and rdnoise is not None:
        nccd = ccdproc.create_deviation(nccd, gain=gain, readnoise=rdnoise)
    elif error and (gain is None or rdnoise is None):
        raise ValueError(
            'gain and rdnoise must be specified to create error frame')

    # apply the bad pixel mask
    if isinstance(bad_pixel_mask, np.ndarray):
        nccd.mask = bad_pixel_mask
    elif bad_pixel_mask is None:
        pass
    else:
        raise TypeError('bad_pixel_mask is not None or numpy.ndarray')

    # apply the gain correction
    if isinstance(gain, u.quantity.Quantity):
        nccd = ccdproc.gain_correct(nccd, gain)
    elif gain is None:
        pass
    else:
        raise TypeError('gain is not None or astropy.Quantity')

    return nccd


def hrs_process(image_name, ampsec=[], oscansec=[], trimsec=[],
                masterbias=None, error=False, bad_pixel_mask=None, flip=False,
                rdnoise=None, oscan_median=True, oscan_model=None):
    """Processing required for HRS observations.  If the images have multiple
       amps, then this will process each part of the image and recombine them
       into for the final results

    Parameters
    ----------
    image_name: str
       Name of file to be processed

    ampsec: list
       List of ampsections.  This list should have the same length as the
       number of amps in the data set.  The sections should be given
       in the format of fits_sections (see below).

    oscansec: list
       List of overscan sections.  This list should have the same length as the
       number of amps in the data set.  The sections should be given
       in the format of fits_sections (see below).

    trimsec: list
       List of overscan sections.  This list should have the same length as the
       number of amps in the data set.  The sections should be given
       in the format of fits_sections (see below).

    error: boolean
        If True, create an uncertainty array for ccd

    masterbias: None, `~numpy.ndarray`,  or `~ccdproc.CCDData`
        A materbias frame to be subtracted from ccd.

    bad_pixel_mask: None or `~numpy.ndarray`
        A bad pixel mask for the data. The bad pixel mask should be in given
        such that bad pixels havea value of 1 and good pixels a value of 0.


    flip: boolean
        If True, the image will be flipped such that the orders run from the
        bottom of the image to the top and the dispersion runs from the left
        to the right.

    rdnoise: None or `~astropy.Quantity`
        Read noise for the observations.  The read noise should be in
        `~astropy.units.electron`

    oscan_median :  bool, optional
        If true, takes the median of each line.  Otherwise, uses the mean

    oscan_model :  `~astropy.modeling.Model`, optional
        Model to fit to the data.  If None, returns the values calculated
        by the median or the mean.

    Returns
     -------
    ccd: `~ccdproc.CCDData`
        Data processed and


    Notes
    -----

    The format of the `fits_section` string follow the rules for slices that
    are consistent with the FITS standard (v3) and IRAF usage of keywords like
    TRIMSEC and BIASSEC. Its indexes are one-based, instead of the
    python-standard zero-based, and the first index is the one that increases
    most rapidly as you move through the array in memory order, opposite the
    python ordering.

    The 'fits_section' argument is provided as a convenience for those who are
    processing files that contain TRIMSEC and BIASSEC. The preferred, more
    pythonic, way of specifying the overscan is to do it by indexing the data
    array directly with the `overscan` argument.

    """
    # read in the data
    ccd = ccdproc.CCDData.read(image_name, unit=u.adu)

    try:
        namps = ccd.header['CCDAMPS']
    except KeyError:
        namps = ccd.header['CCDNAMPS']

    # thow errors for the wrong number of amps
    if len(ampsec) != namps:
        raise ValueError('Number of ampsec does not equal number of amps')
    if len(oscansec) != namps:
        raise ValueError('Number of oscansec does not equal number of amps')
    if len(trimsec) != namps:
        raise ValueError('Number of trimsec does not equal number of amps')

    if namps == 1:
        if ccd.header['OBSTYPE'] == 'Bias':
            gain = None
        else:
            gain = float(ccd.header['gain'].split()[0]) * u.electron / u.adu

        nccd = ccd_process(ccd, oscan=oscansec[0], trim=trimsec[0],
                           error=error, masterbias=masterbias,
                           bad_pixel_mask=bad_pixel_mask, gain=gain,
                           rdnoise=rdnoise, oscan_median=oscan_median,
                           oscan_model=oscan_model)
    else:
        ccd_list = []
        xsize = 0
        ysize = 0
        gain_ave = 0

        # determine size of the image
        ys, xs = ccd.data.shape
        x1 = 0
        x2 = 0
        y1 = 0
        y2 = ys
        for i in range(namps):
            amp = ccdproc.utils.slices.slice_from_string(ampsec[i], fits_convention=True)
            ay1, ay2, _ = amp[0].indices(ys)
            ax1, ax2, _ = amp[1].indices(xs)
            trim = ccdproc.utils.slices.slice_from_string(trimsec[i], fits_convention=True)
            ty1, ty2, _ = trim[0].indices(ys)
            tx1, tx2, _ = trim[1].indices(xs)

            dx1 = max(tx1, ax1)
            x1 = min(x1 or dx1, dx1)

            # assumes bias frames are given in increasing x-order
            x2 = ax2 - (ax2 - ax1 - tx2)

        data = np.zeros((y2 - y1, x2 - x1))
        if ccd.mask is not None:
            mask = np.zeros(((y2 - y1, x2 - x1)))
        else:
            mask = None

        if ccd.uncertainty is not None:
            raise NotImplementedError(
                'Support for uncertainties not implimented yet')
        else:
            uncertainty = None

        for i in range(namps):
            gain_ave += float(ccd.header['gain'].split()[i])
        gain_ave = gain_ave / namps * u.electron / u.adu
        for i in range(namps):
            # trim the image to just the amplifier
            cc = ccdproc.trim_image(ccd, fits_section=ampsec[i])

            # determine its position in the area after correction
            amp = ccdproc.utils.slices.slice_from_string(ampsec[i], fits_convention=True)
            ay1, ay2, _ = amp[0].indices(ys)
            ax1, ax2, _ = amp[1].indices(xs)
            trim = ccdproc.utils.slices.slice_from_string(trimsec[i], fits_convention=True)
            ty1, ty2, _ = trim[0].indices(ys)
            tx1, tx2, _ = trim[1].indices(xs)
            dx1 = max(ax1 - x1, 0)
            dx2 = ax2 - x1
            dy1 = ay1
            dy2 = ay2

            # gain correct the arrray
            gain = float(ccd.header['gain'].split()[i]) * u.electron / u.adu
            ncc = ccd_process(cc, oscan=oscansec[i], trim=trimsec[i],
                              error=False, masterbias=None, gain=gain,
                              bad_pixel_mask=None, rdnoise=rdnoise,
                              oscan_median=oscan_median,
                              oscan_model=oscan_model)

            # normalize all to have the same gain
            ncc = ncc.divide(gain_ave)

            data[dy1:dy2, dx1:dx2] = ncc.data
            if ccd.mask is not None:
                mask[dy1:dy2, dx1:dx2] = ncc.mask

        # create new entryy
        nccd = ccdproc.CCDData(data, unit=ncc.unit, mask=mask,
                               uncertainty=uncertainty)
        nccd.header = ccd.header
        nccd = ccd_process(nccd, masterbias=masterbias, error=error, gain=gain_ave,
                           rdnoise=rdnoise, bad_pixel_mask=bad_pixel_mask)

    if flip:
        nccd.data = nccd.data[::-1, ::-1]
        if (nccd.mask is not None):
            nccd.mask = nccd.mask[::-1, ::-1]
        if (nccd.uncertainty is not None):
            nccd.uncertainty = nccd.uncertainty[::-1, ::-1]

    return nccd


def blue_process(infile, masterbias=None, error=False, rdnoise=None, oscan_correct=False):
    """Process a blue frame
    """
    # check to make sure it is a blue file
    ccd = ccdproc.CCDData.read(infile, unit=u.adu)
    try:
        namps = ccd.header['CCDAMPS']
    except KeyError:
        namps = ccd.header['CCDNAMPS']

    # reduce file
    try:
        blueamp = [ccd.header['AMPSEC'].strip()]
        if oscan_correct:
            bluescan = [ccd.header['BIASSEC'].strip()]
        else:
            bluescan = [None]
        bluetrim = [ccd.header['DATASEC'].strip()]
        # ugly hack for when two amps
        if namps > 1: raise Exception()
    except:
        blueamp = ['[1:1050,:]', '[1051:2100,:]']
        if oscan_correct:
            bluescan = ['[1:26,:]', '[1025:1050,:]']
        else:
            bluescan = [None, None]
        bluetrim = ['[27:1050,:]', '[1:1024,:]']

    flip = True
    ccd = hrs_process(infile, ampsec=blueamp, oscansec=bluescan,
                      trimsec=bluetrim, masterbias=masterbias, error=error,
                      rdnoise=rdnoise, flip=flip)
    # this is in place to deal with changes from one amp to two
    if namps == 1:
        ccd.data = ccd.data[:, ::-1]
        if (ccd.mask is not None):
            ccd.mask = ccd.mask[:, ::-1]
        if (ccd.uncertainty is not None):
            ccd.uncertainty = ccd.uncertainty[:, ::-1]

    return ccd


def red_process(infile, masterbias=None, error=None, rdnoise=None, oscan_correct=False):
    """Process a blue frame
    """

    ccd = ccdproc.CCDData.read(infile, unit=u.adu)
    try:
        namps = ccd.header['CCDAMPS']
    except KeyError:
        namps = ccd.header['CCDNAMPS']

    if namps == 1:
        redamp = ['[1:4122,1:4112]']
        if oscan_correct:
            redscan = ['[1:25,1:4112]']
        else:
            redscan = [None]
        redtrim = ['[27:4122,1:4112]']
    else:
        redamp = ['[1:2074,1:2056]', '[1:2074,2057:4112]', '[2075:4148,1:2056]', '[2075:4122,2057:4112]']
        redscan = namps * [None]
        redtrim = ['[27:2074,:]', '[27:2074,:]', '[1:2048,:]', '[1:2048,:]']

    ccd = hrs_process(infile, ampsec=redamp, oscansec=redscan,
                      trimsec=redtrim, masterbias=masterbias, error=error,
                      rdnoise=rdnoise, flip=False)
    return ccd


def create_masterbias(image_list):
    """Create a master bias frame from a list of images

    Parameters
    ----------
    image_list: list
        List contain the file names to be processed

    Returns
    -------
    masterbias: ccddata.CCDData
        Combine master bias from the biases supplied in image_list

    """
    # determine whether they are red or blue
    if os.path.basename(image_list[0]).startswith('H'):
        func = blue_process
    elif os.path.basename(image_list[0]).startswith('R'):
        func = red_process
    else:
        raise TypeError('These are not standard HRS frames')

    # reduce the files
    ccd_list = []
    for image_name in image_list:
        ccd = func(image_name, masterbias=None, error=False)
        ccd_list.append(ccd)

    # combine the files
    cb = ccdproc.Combiner(ccd_list)
    nccd = cb.median_combine(median_func=np.median)

    return nccd


def flatfield_science_order(hrs, flat_hrs, median_filter_size=None, interp=False):
    """Apply a flat field to an hrs order.  If median_filter_size is set,
       the process first removes the overall flat shape by dividing out
       a median filter of the data.   Then it is divided through the
       science data.  This is done on a row by row basis after removing
       the shape of the science data.

    Parameters
    ----------

    hrs: `~pyhrs.HRSOrder`
        Object describing a single HRS order

    flat_hrs: `~pyhrs.HRSOrder`
        Object describing the flatfield for a single HRS order

    median_filter_size: None or int
        Size for median filter to be run on the data and remove the general
        flat field shape

    Returns
    -------
    hrs: `~pyhrs.HRSOrder`
        An hrs object with the flatfield applied to the flux property.

    """

    # set through each line and correct for the shape of the flux
    # if needed
    if median_filter_size is not None:
        # set up the box for the flat hrs
        fbox, coef = flat_hrs.create_box(flat_hrs.flux, interp=interp)
        # for each line, correct for the shape of the flux
        y1 = 0
        y2 = len(fbox)
        for i in range(y1, y2):
            if np.any(fbox[i] > 0):
                fbox[i][fbox[i] == np.inf] = 0
                sf = nd.filters.median_filter(fbox[i], size=median_filter_size)
                sf[sf <= 0] = max(sf.mean(), 1)
                fbox[i] = fbox[i] / sf

        # return to the original shape
        flat_hrs.flux = flat_hrs.unravel_box(fbox)

    # step to prevent divide by zero
    flat_hrs.flux[flat_hrs.flux == 0] = 1

    # divide and normalize by the flux
    hrs.flux = hrs.flux * np.median(flat_hrs.flux) / flat_hrs.flux

    if hrs.error is not None:
        hrs.error = hrs.error * np.median(flat_hrs.flux) / flat_hrs.flux

    return hrs


def flatfield_science(ccd, flat_frame, order_frame, median_filter_size=None, interp=False):
    """Flatfield all of the orders in a science frame

    Parameters
    ----------
    ccd: ~ccdproc.CCDData
        Science frame to be flatfielded

    flar_frame: ~ccdproc.CCDData
        Frame containing the flat field for each of the orders

    order_frame: ~ccdproc.CCDData
        Frame containting the positions of each of the orders

     median_filter_size: None or int
        Size for median filter to be run on the data and remove the general
        flat field shape

    Returns
    -------
    ccd: ~ccdproc.CCDData
        Flatfielded science frame

    """
    # get a list of orders
    o1 = order_frame.data[order_frame.data > 0].min()
    o2 = order_frame.data.max()
    order_arr = np.arange(o1, o2, dtype=int)

    ndata = 0.0 * ccd.data
    if ccd.uncertainty is not None:
        edata = ccd.uncertainty.array
    else:
        edata = None

    for n_order in order_arr:
        hrs = HRSOrder(n_order)
        hrs.set_order_from_array(order_frame.data)
        hrs.set_flux_from_array(ccd.data, flux_unit=ccd.unit, error=edata)

        flat_hrs = HRSOrder(n_order)
        flat_hrs.set_order_from_array(order_frame.data)
        flat_hrs.set_flux_from_array(flat_frame.data, flux_unit=flat_frame.unit)

        hrs = flatfield_science_order(hrs, flat_hrs, median_filter_size=median_filter_size, interp=interp)

        ndata[hrs.region[0], hrs.region[1]] = hrs.flux
        if edata is not None: edata[hrs.region[0], hrs.region[1]] = hrs.error

    ccd.data = ndata
    return ccd



from __future__ import absolute_import, division, print_function

from signal import SIGINT, signal as sigsignal
from warnings import resetwarnings, simplefilter
from time import time
from pypit import armsgs
from pypit import archeck
import glob
import numpy as np

# Import PYPIT routines

try:
    from linetools.spectra.xspectrum1d import XSpectrum1D
except ImportError:
    pass

from pypit import ardebug as debugger


def PYPIT(redname, debug=None, progname=__file__, quick=False, ncpus=1, verbosity=1,
          use_masters=False, devtest=False, logname=None):
    """ Main driver of the PYPIT code. Default settings and
    user-specified changes are made, and passed to the
    appropriate code for data reduction.

    Parameters
    ----------
    redname : string
      Input reduction script
    debug : dict, optional
      Debug dict
    progname : string
      Name of the program
    quick : bool
      If True, a quick reduction (but possibly less
      accurate) will be performed. This flag is most
      useful for observing at a telescope, but not
      for publication quality results.
    ncpus : int
      Number of CPUs to use for multiprocessing the
      data reduction (sometimes not used)
    verbosity : int (0,1,2)
      Level of verbosity:
        0 = No output
        1 = Minimal output (default - suitable for the average user)
        2 = All output
    use_masters : bool, optional
      Load calibration files from MasterFrames directory, if they exist
    devtest : bool, optional
      Running PYPIT-Development suite;  will turn instrument-specific options
    logname : str or None
          The name of an ascii log file which is used to
          save the output details of the reduction
        debug : dict
          A PYPIT debug dict (from ardebug.init)
        version : str
        last_updated : str
    ---------------------------------------------------
    """
    from pypit import ardebug
    # Init logger
    if debug is None:
        debug = ardebug.init()
    msgs = armsgs.get_logger((logname, debug, verbosity))
    msgs.pypit_file = redname

    # This needs to be loaded after msgs
    from pypit import arparse

    # version checking
    try:
        archeck.version_check()
    except archeck.VersionError as err:
        msgs.error(err.message)

    # First send all signals to messages to be dealt with (i.e. someone hits ctrl+c)
    sigsignal(SIGINT, msgs.signal_handler)

    # Ignore all warnings given by python
    resetwarnings()
    simplefilter("ignore")

    # Record the starting time
    tstart = time()

    # Load the input file
    pyp_dict = load_input(redname, msgs)
    parlines, datlines, spclines = [pyp_dict[ii] for ii in ['par','dat','spc']]

    # Initialize the arguments and flags
#    argflag = arload.argflag_init()
#    settings.argflag['run']['ncpus'] = ncpus
#    settings.argflag['output']['verbosity'] = verbosity

    # Determine the name of the spectrograph
    specname = None
    for i in range(len(parlines)):
        parspl = parlines[i].split()
        if len(parspl) < 3:
            msgs.error("There appears to be a missing argument on the following input line" + msgs.newline() +
                       parlines[i])
        if (parspl[0] == 'run') and (parspl[1] == 'spectrograph'):
            specname = parspl[2]
            break
    if specname is None:
        msgs.error("Please specify the spectrograph settings to be used with the command" + msgs.newline() +
                   "run spectrograph <name>")
    msgs.info("Reducing data from the {0:s} spectrograph".format(specname))

    # Determine the type of reduction used for this spectrograph
    redtype = None
    # Get the software path
    prgn_spl = progname.split('/')
    tfname = "/".join(prgn_spl[:-1]) + "/"
    # Settings file
    fname = tfname + 'data/settings/settings.' + specname
    try:
        spl = open(fname, 'r').readlines()
    except IOError:
        msgs.error("The following instrument settings file cannot be found:" + msgs.newline() + fname + msgs.newline() +
                   "Please check the settings file exists, and that the instrument name is spelt correctly.")
    for i in range(len(spl)):
        parspl = spl[i].split()
        if len(parspl) < 3:
            continue
        if (parspl[0] == 'mosaic') and (parspl[1] == 'reduction'):
            redtype = parspl[2]
            break
    if redtype is None:
        msgs.bug("The {0:s} instrument settings file must contain the reduction type".format(specname))
        msgs.error("Please specify the reduction type with the command" + msgs.newline() +
                   "mosaic reduction <type>")

    # Load default reduction arguments/flags, and set any command line arguments
    argf = arparse.get_argflag_class((redtype.upper(), ".".join(redname.split(".")[:-1])))
    argf.init_param()
    # Run specific
    argf.set_param('run pypitdir {0:s}'.format(tfname))
    argf.set_param('run progname {0:s}'.format(progname))
    argf.set_param('run redname {0:s}'.format(redname))
    # Load user changes to the arguments/flags
    plines = argf.load_lines(parlines)
    argf.set_paramlist(plines)
    # If the user wishes to load a settings file, do that now
    if argf.__dict__['_argflag']['run']['load']['settings'] is not None:
        lines = argf.load_file(argf.__dict__['_argflag']['run']['load']['settings'])
        argf.set_paramlist(lines)

    # Load default spectrograph settings
    spect = arparse.get_spect_class((redtype.upper(), specname, ".".join(redname.split(".")[:-1])))
    lines = spect.load_file(base=True)  # Base spectrograph settings
    spect.set_paramlist(lines)
    lines = spect.load_file()  # Instrument specific
    spect.set_paramlist(lines)
    # Load frametype numbers, as relevant
    if len(pyp_dict['ftype']) > 0:
        ftlines = spect.load_ftype(pyp_dict['ftype'])
        plines = spect.load_lines(ftlines)
        spect.set_paramlist(plines)
    # Load user changes to the arguments/flags
    plines = spect.load_lines(spclines)
    spect.set_paramlist(plines)
    if argf.__dict__['_argflag']['run']['load']['spect'] is not None:
        lines = spect.load_file(argf.__dict__['_argflag']['run']['load']['spect'])
        spect.set_paramlist(lines)
    # If the instrument settings file sets some argflag settings, implement those changes now
    if len(spect.__dict__['_settings']) != 0:
        argf.set_paramlist(spect.__dict__['_settings'])
    # Load command line changes
    argf.set_param('run ncpus {0:d}'.format(ncpus))
    argf.set_param('output verbosity {0:d}'.format(verbosity))
    if use_masters:
        argf.set_param('reduce masters reuse True')
    msgs.work("Make appropriate changes to quick reduction")
    # Load Development suite changes
    if devtest:
        msgs.info("Loading instrument specific argurment for Development Suite tests")
        from pypit import ardevtest
        ardevtest.set_param(argf, specname)

    if quick:
        # If a quick reduction has been requested, make sure the requested pipeline
        # is the quick implementation (if it exists), otherwise run the standard pipeline.
        msgs.work("QUICK REDUCTION TO STILL BE DONE")
    # Setup from PYPIT file?
    if len(pyp_dict['setup']['name']) == 1:
        argf.set_param('setup name {:s}'.format(pyp_dict['setup']['name'][0]))
    # Finally, save the arguments/flags and spectrograph settings used for this reduction
    argf.save()
    spect.save()

    # Now that all of the relevant settings are loaded, globalize the settings
    arparse.init(argf, spect)

    '''
    # Test that a maximum of one .setup files is present
    from pypit import arsort
    setup_file, nexist = arsort.get_setup_file()
    if nexist == 1:
        msgs.info("Found setup_file: {:s}".format(setup_file))
        msgs.info("Will use this to guide the data reduction.")
    '''

    # Load the important information from the fits headers
    from pypit.arload import load_headers
    fitsdict, updates = load_headers(datlines)

    # If some settings were updated because of the fits headers, globalize the settings again
    if len(updates) != 0:
        spect.set_paramlist(updates)
        arparse.init(argf, spect)

    # If the dispersion direction is 1, flip the axes
    if arparse.argflag['trace']['dispersion']['direction'] == 1:
        # Update the keywords of all fits files
        for ff in range(len(fitsdict['naxis0'])):
            temp = fitsdict['naxis0'][ff]
            fitsdict['naxis0'][ff] = fitsdict['naxis1'][ff]
            fitsdict['naxis1'][ff] = temp
        # Update the spectrograph settings for all detectors in the mosaic
        for dd in range(arparse.spect['mosaic']['ndet']):
            ddnum = arparse.get_dnum(dd+1)
            # Change the user-specified (x,y) pixel sizes
            tmp = arparse.spect[ddnum]['xgap']
            arparse.spect[ddnum]['xgap'] = arparse.spect[ddnum]['ygap']
            arparse.spect[ddnum]['ygap'] = tmp
            arparse.spect[ddnum]['ysize'] = 1.0 / arparse.spect[ddnum]['ysize']
            # Update the amplifier/data/overscan sections
            for i in range(arparse.spect[ddnum]['numamplifiers']):
                # Flip the order of the sections
                arparse.spect[ddnum]['datasec{0:02d}'.format(i + 1)] = arparse.spect[ddnum][
                                                                            'datasec{0:02d}'.format(i + 1)][::-1]
                arparse.spect[ddnum]['oscansec{0:02d}'.format(i + 1)] = arparse.spect[ddnum][
                                                                             'oscansec{0:02d}'.format(i + 1)][::-1]
    # Reduce the data!
    status = 0
    # Send the data away to be reduced
    if spect.__dict__['_spect']['mosaic']['reduction'] == 'ARMLSD':
        msgs.info("Data reduction will be performed using PYPIT-ARMLSD")
        from pypit import armlsd
        status = armlsd.ARMLSD(fitsdict)
    elif spect.__dict__['_spect']['mosaic']['reduction'] == 'ARMED':
        msgs.info("Data reduction will be performed using PYPIT-ARMED")
        from pypit import armed
        status = armed.ARMED(fitsdict)
    # Check for successful reduction
    if status == 0:
        from pypit import arqa
        msgs.info("Data reduction complete")
        # QA HTML
        msgs.info("Generating QA HTML")
        arqa.gen_mf_html(redname)
        arqa.gen_exp_html()
    elif status == 1:
        msgs.info("Setup complete")
    elif status == 2:
        msgs.info("Calcheck complete")
    else:
        msgs.error("Data reduction failed with status ID {0:d}".format(status))
    # Capture the end time and print it to user
    tend = time()
    codetime = tend-tstart
    if codetime < 60.0:
        msgs.info("Data reduction execution time: {0:.2f}s".format(codetime))
    elif codetime/60.0 < 60.0:
        mns = int(codetime/60.0)
        scs = codetime - 60.0*mns
        msgs.info("Data reduction execution time: {0:d}m {1:.2f}s".format(mns, scs))
    else:
        hrs = int(codetime/3600.0)
        mns = int(60.0*(codetime/3600.0 - hrs))
        scs = codetime - 60.0*mns - 3600.0*hrs
        msgs.info("Data reduction execution time: {0:d}h {1:d}m {2:.2f}s".format(hrs, mns, scs))
    return


def load_input(redname, msgs):
    """
    Load user defined input .pypit reduction file. Updates are
    made to the argflag dictionary.

    Parameters
    ----------
    redname : string
      Name of reduction script
    msgs : Messages
      logger for PYPIT

    Returns
    -------
    pyp_dict : dict
      Contains the following keys --
      'par'
        parlines : list
          Input (uncommented) lines specified by the user.
          parlines is used in this routine to update the
          argflag dictionary
      'dat'
        datlines : list
          Input (uncommented) lines specified by the user.
          datlines contains the full data path to every
          raw exposure listed by the user
      'spc'
        spclines : list
          Input (uncommented) lines specified by the user.
          spclines contains a list of user-specified changes
          that should be made to the default spectrograph
          settings.
      'dfn'
        dfnames : list
          Input data lines
      'setup'
        dict of setup info
          'name' list of setups
          'lines' list of lines in the setup block
      'ftype' : dict
         dict of filename: frametype
    """
    import os
    # Read in the model file
    msgs.info("Loading the input file")
    try:
        infile = open(redname, 'r')
    except IOError:
        msgs.error("The filename does not exist -" + msgs.newline() + redname)
    lines = infile.readlines()
    parlines = []
    datlines = []
    skip_files = []
    spclines = []
    dfnames = []
    setuplines = []
    paths = []
    rddata, rdspec, rdsetup, rdsfiles = 0, 0, 0, -1
    setups = []
    ftype_dict = {}
    ftype_col = -1
    for i in range(len(lines)):
        if lines[i].strip() == '': continue
        linspl = lines[i].split()
        if rddata == 1: # Read datafile(s)
            if linspl[0] == 'data' and linspl[1] == 'end':
                rddata += 1
                # Deal with skip files
                if len(skip_files) > 0:
                    keep = np.array([True]*len(datlines))
                    for skip_file in skip_files:
                        for kk,datfile in enumerate(datlines):
                            if skip_file in datfile:
                                keep[kk] = False
                                msgs.warn("Skipping file {:s}".format(skip_file))
                    # Save
                    datlines = np.array(datlines)[keep].tolist()
                continue
            #
            dfname = lines[i].rstrip('\n').strip()
            if rdsfiles == -1:
                if 'path' in dfname[0:5]:
                    rdsfiles = 1
                else:
                    rdsfiles = 0
            if rdsfiles == 0:
                # is there a comment?
                aux = dfname.split('#')
                if len(aux) > 1:  # yes, there is a comment
                    dfname = aux[0].strip()
                if len(dfname) == 0:  # line is fully commented out
                    continue
                elif dfname[0] == '~':
                    dfname = os.path.expanduser(dfname)
                    print(dfname)
                elif dfname[:4] == 'skip':
                    skip_files.append(dfname.split(' ')[1])
                elif dfname[0] != '/':
                    msgs.error("You must specify the full datapath for the file:" + msgs.newline() + dfname)
                elif len(dfname.split()) != 1:
                    msgs.error("There must be no spaces when specifying the datafile:" + msgs.newline() + dfname)
                dfnames.append(dfname)
                listing = glob.glob(dfname)
                for lst in listing: datlines.append(lst)
            else:  # File by file approach
                if 'path' in dfname[0:5]:
                    paths.append(linspl[1])
                else:  # Grab filename and frametype
                    if ftype_col == -1:  # Identify columns for frametype
                        ftype_col = np.where(np.array(linspl) == 'frametype')[0][0]
                        dfile_col = np.where(np.array(linspl) == 'filename')[0][0]
                    else:
                        # Skip commented lines
                        if lines[i][0] == '#':
                            continue
                        # Find datafile using last used path and update ftype dict
                        path = paths[-1]
                        if os.path.isfile(path+linspl[dfile_col]):
                            datlines.append(path+linspl[dfile_col])
                            ftype_dict[linspl[dfile_col]] = linspl[ftype_col]
            continue
        elif rddata == 0 and linspl[0] == 'data' and linspl[1] == 'read': # Begin data read block
            rddata += 1
            continue
        if rdsetup == 1:  # Read setup command
            if linspl[0] == 'setup' and linspl[1] == 'end':
                rdsetup += 1
                continue
            if 'Setup' in lines[i]:
                setups.append(lines[i][6:].strip())
            setuplines.append(lines[i])
            continue
        elif rdsetup == 0 and linspl[0] == 'setup' and linspl[1] == 'read':  # Begin setup read block
            rdsetup += 1
            continue
        if rdspec == 1:  # Read spect command
            if linspl[0] == 'spect' and linspl[1] == 'end':
                rdspec += 1
                continue
            spclines.append(lines[i])
            continue
        elif rdspec == 0 and linspl[0] == 'spect' and linspl[1] == 'read':  # Begin spect read block
            rdspec += 1
            continue
        if lines[i].lstrip()[0] == '#': continue
        parlines.append(lines[i])
    # Do some quick checks
    if rddata == 0:
        msgs.error("You haven't specified any data!")
    elif rddata == 1:
        msgs.error("Missing 'data end' in " + redname)
    if rddata == 0:
        msgs.info("Using Default spectrograph parameters")
    elif rddata != 2:
        msgs.error("Missing 'spect end' in " + redname)
    # Check there are no duplicate inputs
    if len(datlines) != len(set(datlines)):
        msgs.error("There are duplicate files in the list of data.")
    if len(datlines) == 0:
        msgs.error("There are no raw data frames" + msgs.newline() +
                   "Perhaps the path to the data is incorrect?")
    else:
        msgs.info("Found {0:d} raw data frames".format(len(datlines)))
    msgs.info("Input file loaded successfully")
    # Let's return a dict
    pypit_dict = dict(par=parlines, dat=datlines, spc=spclines,
                      dfn=dfnames, setup={'name': setups, 'lines': setuplines},
                    ftype=ftype_dict)
    return pypit_dict # parlines, datlines, spclines, dfnames, setup, setuplines, ftype_dict



import sys
import os
import numpy as np

from astropy.io import fits
from astropy import units as u

import ccdproc
from ccdproc import CCDData

from ccdproc import ImageFileCollection

if len(sys.argv) != 3:
    print('Usage:\npython wht_basic_rection.py [full_path_to_raw_data] [full_path_to_reduced_data]\n')
    exit()

indir = sys.argv[1]
outdir = sys.argv[2]

if not os.path.isdir(outdir): os.mkdir(outdir)
os.chdir(outdir)

# change this to point to your raw data directory
ic1 = ImageFileCollection(indir)

# create the bias frames
blue_bias_list = []
for filename in ic1.files_filtered(obstype='Bias', isiarm='Blue arm'):
    print ic1.location + filename
    ccd = CCDData.read(ic1.location + filename, unit=u.adu)
    # this has to be fixed as the bias section does not include the whole section that will be trimmed
    ccd = ccdproc.subtract_overscan(ccd, median=True, overscan_axis=0, fits_section='[1:966,4105:4190]')
    ccd = ccdproc.trim_image(ccd, fits_section=ccd.header['TRIMSEC'])
    blue_bias_list.append(ccd)
master_bias_blue = ccdproc.combine(blue_bias_list, method='median')
master_bias_blue.write('master_bias_blue.fits', clobber=True)

red_bias_list = []
for filename in ic1.files_filtered(obstype='Bias', isiarm='Red arm'):
    print ic1.location + filename
    ccd = CCDData.read(ic1.location + filename, unit=u.adu)
    # this has to be fixed as the bias section does not include the whole section that will be trimmed
    ccd = ccdproc.subtract_overscan(ccd, median=True, overscan_axis=0, fits_section='[1:966,4105:4190]')
    ccd = ccdproc.trim_image(ccd, fits_section=ccd.header['TRIMSEC'])
    red_bias_list.append(ccd)
master_bias_red = ccdproc.combine(red_bias_list, method='median')
master_bias_red.write('master_bias_red.fits', clobber=True)

# create the flat fields
red_flat_list = []
for filename in ic1.files_filtered(obstype='Flat', isiarm='Red arm'):
    ccd = CCDData.read(ic1.location + filename, unit=u.adu)
    # this has to be fixed as the bias section does not include the whole section that will be trimmed
    ccd = ccdproc.subtract_overscan(ccd, median=True, overscan_axis=0, fits_section='[1:966,4105:4190]')
    ccd = ccdproc.trim_image(ccd, fits_section=ccd.header['TRIMSEC'])
    ccd = ccdproc.subtract_bias(ccd, master_bias_red)
    red_flat_list.append(ccd)
master_flat_red = ccdproc.combine(red_flat_list, method='median')
master_flat_red.write('master_flat_red.fits', clobber=True)

blue_flat_list = []
for filename in ic1.files_filtered(obstype='Flat', isiarm='Blue arm'):
    ccd = CCDData.read(ic1.location + filename, unit=u.adu)
    # this has to be fixed as the bias section does not include the whole section that will be trimmed
    ccd = ccdproc.subtract_overscan(ccd, median=True, overscan_axis=0, fits_section='[1:966,4105:4190]')
    ccd = ccdproc.trim_image(ccd, fits_section=ccd.header['TRIMSEC'])
    ccd = ccdproc.subtract_bias(ccd, master_bias_blue)
    blue_flat_list.append(ccd)
master_flat_blue = ccdproc.combine(blue_flat_list, method='median')
master_flat_blue.write('master_flat_blue.fits', clobber=True)

# reduce the arc frames
for filename in ic1.files_filtered(obstype='Arc', isiarm='Blue arm'):
    hdu = fits.open(ic1.location + filename)
    ccd = CCDData(hdu[1].data, header=hdu[0].header + hdu[1].header, unit=u.adu)
    # this has to be fixed as the bias section does not include the whole section that will be trimmed
    ccd = ccdproc.subtract_overscan(ccd, median=True, overscan_axis=0, fits_section='[1:966,4105:4190]')
    ccd = ccdproc.trim_image(ccd, fits_section=ccd.header['TRIMSEC'])
    ccd = ccdproc.subtract_bias(ccd, master_bias_blue)
    ccd = ccdproc.flat_correct(ccd, master_flat_blue)
    ccd.data = ccd.data.T
    ccd.write('arc_' + filename, clobber=True)

red_flat_list = []
for filename in ic1.files_filtered(obstype='Arc', isiarm='Red arm'):
    hdu = fits.open(ic1.location + filename)
    ccd = CCDData(hdu[1].data, header=hdu[0].header + hdu[1].header, unit=u.adu)
    # this has to be fixed as the bias section does not include the whole section that will be trimmed
    ccd = ccdproc.subtract_overscan(ccd, median=True, overscan_axis=0, fits_section='[1:966,4105:4190]')
    ccd = ccdproc.trim_image(ccd, fits_section=ccd.header['TRIMSEC'])
    ccd = ccdproc.subtract_bias(ccd, master_bias_red)
    ccd = ccdproc.flat_correct(ccd, master_flat_red)
    ccd.data = ccd.data.T
    ccd.write('arc_' + filename, clobber=True)

# reduce the sky frames
for filename in ic1.files_filtered(obstype='Sky', isiarm='Blue arm'):
    hdu = fits.open(ic1.location + filename)
    ccd = CCDData(hdu[1].data, header=hdu[0].header + hdu[1].header, unit=u.adu)
    # this has to be fixed as the bias section does not include the whole section that will be trimmed
    ccd = ccdproc.subtract_overscan(ccd, median=True, overscan_axis=0, fits_section='[1:966,4105:4190]')
    ccd = ccdproc.trim_image(ccd, fits_section=ccd.header['TRIMSEC'])
    ccd = ccdproc.subtract_bias(ccd, master_bias_blue)
    ccd = ccdproc.flat_correct(ccd, master_flat_blue)
    ccd.data = ccd.data.T
    ccd.write('sky_' + filename, clobber=True)

for filename in ic1.files_filtered(obstype='Sky', isiarm='Red arm'):
    hdu = fits.open(ic1.location + filename)
    ccd = CCDData(hdu[1].data, header=hdu[0].header + hdu[1].header, unit=u.adu)
    # this has to be fixed as the bias section does not include the whole section that will be trimmed
    ccd = ccdproc.subtract_overscan(ccd, median=True, overscan_axis=0, fits_section='[1:966,4105:4190]')
    ccd = ccdproc.trim_image(ccd, fits_section=ccd.header['TRIMSEC'])
    ccd = ccdproc.subtract_bias(ccd, master_bias_red)
    ccd = ccdproc.flat_correct(ccd, master_flat_red)
    ccd.data = ccd.data.T
    ccd.write('sky_' + filename, clobber=True)

# reduce the object frames
for filename in ic1.files_filtered(obstype='TARGET', isiarm='Blue arm'):
    hdu = fits.open(ic1.location + filename)
    ccd = CCDData(hdu[1].data, header=hdu[0].header + hdu[1].header, unit=u.adu)
    # this has to be fixed as the bias section does not include the whole section that will be trimmed
    ccd = ccdproc.subtract_overscan(ccd, median=True, overscan_axis=0, fits_section='[1:966,4105:4190]')
    ccd = ccdproc.trim_image(ccd, fits_section=ccd.header['TRIMSEC'])
    ccd = ccdproc.subtract_bias(ccd, master_bias_blue)
    ccd = ccdproc.flat_correct(ccd, master_flat_blue)
    ccd.data = ccd.data.T
    ccd.write('obj_' + filename, clobber=True)

for filename in ic1.files_filtered(obstype='Target', isiarm='Red arm'):
    hdu = fits.open(ic1.location + filename)
    ccd = CCDData(hdu[1].data, header=hdu[0].header + hdu[1].header, unit=u.adu)
    # this has to be fixed as the bias section does not include the whole section that will be trimmed
    ccd = ccdproc.subtract_overscan(ccd, median=True, overscan_axis=0, fits_section='[1:966,4105:4190]')
    ccd = ccdproc.trim_image(ccd, fits_section=ccd.header['TRIMSEC'])
    ccd = ccdproc.subtract_bias(ccd, master_bias_red)
    ccd = ccdproc.flat_correct(ccd, master_flat_red)
    ccd.data = ccd.data.T
    ccd.write('obj_' + filename, clobber=True)


import os
import numpy as np
from subprocess import Popen, DEVNULL
import shlex
from glob import glob
from astropy.io import fits
from ccdproc import  ImageFileCollection
import astropy.io.votable as votable


# TODO: Offer an iterative treatment of SCAMP to get desired precision
# TODO: Photometric calibration / selection of filters

def go_sextractor(inputfile,
                configfile=None,
                paramfile = None,
                convfile = 'default.conv',
                nnwfile = 'default.nnw',
                verbose=True, clean=True):
    """
    """
    # path = os.path.abspath(amodule.__file__)

    if configfile == None:
        configfile = 'sextractor.lbc.conf'

    # TODO: Use default if output.param file doesnt exist
    if paramfile == None:
        paramfile = 'sextractor.lbcoutput.param'
        file_exists = os.path.exists(paramfile)
        if not file_exists:
            print("Warning: SEXTRACTOR paramater file {0} "
                  "does not exist".format(paramfile))
            return None

    # Test for convolution file
    file_exists = os.path.exists(convfile)
    if not file_exists:
        print("Warning: SEXTRACTOR convolution file {0} "
              "does not exist".format(convfile))
        return None

    # Test for NNW file
    file_exists = os.path.exists(nnwfile)
    if not file_exists:
        print("Warning: SEXTRACTOR convolution file {0} "
              "does not exist".format(nnwfile))
        return None

    # Base filename:
    filebase = inputfile.replace('.fits','')

    # Source extractor catalog suffix
    outputsuffix = '.cat'
    outputcatalog = filebase+outputsuffix

    # SExtractor flags
    cmd_flags = ' -c '+ configfile + \
        ' -CATALOG_NAME '+outputcatalog + \
        ' -CATALOG_TYPE FITS_LDAC'+ \
        ' -DETECT_THRESH 5.0 -ANALYSIS_THRESH 8.0'+ \
        ' -PARAMETERS_NAME '+paramfile

    # Put together the SExtractor command
    cmd = 'sex '+inputfile+cmd_flags

    try:
        if verbose:
            print('########### SEXTRACTOR run for {0} '
                  '########### \n'.format(inputfile.replace('.cat', '')))
            print(cmd)
            sextract = Popen(shlex.split(cmd),
                             close_fds=True)
        else:
            sextract = Popen(shlex.split(cmd),
                             stdout=DEVNULL,
                             stderr=DEVNULL,
                             close_fds=True)
    except Exception as e:
        print('Oops: source extractor call:', (e))
        return None

    sextract.wait()


def go_scamp(inputfile,
             astrometric_catalog='GAIA-DR2',
             astrometric_method = 'exposure',
             num_iterations = 3,
             configfile=None,
             verbose=True, clean=True):

    """ Run Astromatic.net code SCAMP to calculate astrometric
     solution on an image.
    """

    # Make sure the input file is a SEXTRACTOR catalog:
    inputfile = inputfile.replace('.fits','.cat')
    xmlfile = inputfile.replace('.cat','.xml')

    # Make sure we have a configuration file:
    if configfile == None:
        # TODO: replace SCAMP config file with default config file for LBCgo.
        configfile = 'scamp.lbc.conf'

    # Using only a single iteration of SCAMP doesn't do well enough. Force at
    # least two iterations:
    if num_iterations < 2:
        print('WARNING: Use at least 2 SCAMP iterations. Setting num_iterations = 2...')
        num_iterations = 2

    # Perform the iterations
    for scmpiter in np.arange(num_iterations):
        if scmpiter == 0:
            mosaic_type = 'LOOSE'
            pixscale_maxerr = '1.2'
            position_maxerr = '1'
            posangle_maxerr = '5.0'
            crossid_radius = '7.5'
            aheader_suffix = '.ahead'
        elif scmpiter == 1:
           mosaic_type = 'FIX_FOCALPLANE'
           pixscale_maxerr = '1.1'
           position_maxerr = '0.1'
           posangle_maxerr = '3.0'
           crossid_radius = '5.0'
           aheader_suffix = '.head'
        elif scmpiter == 2:
           mosaic_type = 'FIX_FOCALPLANE'
           pixscale_maxerr = '1.05'
           position_maxerr = '0.05'
           posangle_maxerr = '1.0'
           crossid_radius = '5'
           aheader_suffix = '.head'
        else:
           mosaic_type = 'FIX_FOCALPLANE'
           pixscale_maxerr = '1.05'
           position_maxerr = '0.025'
           posangle_maxerr = '1.0'
           crossid_radius = '2.5'
           aheader_suffix = '.head'

        cmd_flags = ' -c '+ configfile + \
            ' -PIXSCALE_MAXERR '+pixscale_maxerr+ \
            ' -POSANGLE_MAXERR '+posangle_maxerr+ \
            ' -POSITION_MAXERR '+position_maxerr+ \
            ' -ASTREF_CATALOG '+astrometric_catalog+ \
            ' -AHEADER_SUFFIX '+aheader_suffix+ \
            ' -CROSSID_RADIUS '+crossid_radius+\
            ' -XML_NAME '+xmlfile
        # ' -MOSAIC_TYPE '+mosaic_type+ \

        if astrometric_method == 'exposure':
            cmd_flags.replace('INSTRUMENT','EXPOSURE')

        # Create the final command:
        cmd = 'scamp '+inputfile+cmd_flags

        try:
            if verbose:
                print('########### SCAMP iteration {0} for {1} '
                '########### \n'.format(scmpiter+1,
                inputfile.replace('.cat','')))
                # Diagnostics
                print(cmd)

                scamp = Popen(shlex.split(cmd),
                                   close_fds=True)
            else:
                scamp = Popen(shlex.split(cmd),
                                   stdout=DEVNULL,
                                   stderr=DEVNULL,
                                   close_fds=True)
        except Exception as e:
            print('Oops: source Extractor call:', (e))
            return None

        scamp.wait()


    # Read XML file after last iteration
    scamp_diagnostic = (votable.parse(xmlfile)).get_first_table().array
    xy_dispersion = scamp_diagnostic['AstromSigma_Reference'].data
    astrometric_dispersion = np.sqrt(np.sum(xy_dispersion**2))

    # TODO: Do something with the astrometric dispersion

    if clean:
        # Clean the GAIA catalog files
        catfiles = glob('GAIA*cat')
        for ctfls in catfiles:
            cmd = 'rm '+ctfls
            clean_dir = Popen(shlex.split(cmd),
                close_fds=True)
            clean_dir.wait()



def go_swarp(inputfiles,
             output_filename = None,
             configfile = None,
             verbose = True):
    """Do SWARP"""

    # Make sure we have a configuration file:
    if configfile == None:
        # TODO: replace this with default config file in LBCgo directories.
        configfile = 'swarp.lbc.conf'

    # Gather some information about the input files
    keywds = ['object', 'filter', 'exptime', 'imagetyp', 'propid', 'lbcobnam',
                  'airmass', 'HA', 'objra', 'objdec']
    ic_swarp = ImageFileCollection('./', keywords=keywds,
                                    filenames = inputfiles)

    # Check all are the same filter:
    # TODO: Set exception if not all are same filter rather than just bail.
    fltrs = ic_swarp.values('filter',unique=True)
    if np.size(fltrs) != 1:
        print('Warning: not all files in SWARP call use the same filter.')
        return None

    # Calculate mean airmass (weighted by exposure time)
    exp_airmass = np.array(ic_swarp.values('airmass'))
    exp_time = np.array(ic_swarp.values('exptime'))
    airmass = np.average(exp_airmass,weights=exp_time)

    # from IPython import embed ; embed()

    # For now grab the information from the first header:
    if output_filename == None:
        imhead = fits.getheader(inputfiles[0])
        # Shorten the filter names used:
        filter_text = imhead['FILTER']
        filter_text = filter_text.replace('-SLOAN','').replace('-BESSEL','').replace('SDT_Uspec','Uspec')
        # Create final output filename
        # TODO: Check if file exists.
        output_filename = (imhead['object']).\
            replace(' ','')+'.'+filter_text+'.mos.fits'

    # Rename the weight image
    weight_filename = output_filename.replace('.mos.fits','.mos.weight.fits')

    # Create the list of input files:
    inputfile_text = ''
    for fl in inputfiles: inputfile_text = inputfile_text+' '+fl

    cmd_flags = ' -c '+configfile+ \
        ' -IMAGEOUT_NAME '+ output_filename + \
        ' -WEIGHTOUT_NAME '+ weight_filename + \
        ' -WEIGHT_TYPE NONE '+ \
        ' -HEADER_SUFFIX ".head"'+ \
        ' -FSCALE_KEYWORD NONE -FSCALE_DEFAULT 1.0 '+\
        ' -CELESTIAL_TYPE EQUATORIAL -CENTER_TYPE ALL '+\
        ' -COMBINE_BUFSIZE 4096 ' +\
        ' -COPY_KEYWORDS '+\
        ' OBJECT,OBJRA,OBJDEC,OBJEPOCH,PROPID,PI_NAME,'+\
        'FILTER,SATURATE,RDNOISE,GAIN,EXPTIME,AIRMASS,TIME-OBS'

    # Create the final command:
    cmd = 'swarp ' + inputfile_text + cmd_flags

    try:
        if verbose:
            print('########### SWARP image combination '
                  '########### \n')
            print(cmd)
            swarp = Popen(shlex.split(cmd),
                          close_fds=True)
        else:
            swarp = Popen(shlex.split(cmd),
                          stdout=DEVNULL,
                          stderr=DEVNULL,
                          close_fds=True)
    except Exception as e:
        print('Oops: SWARP call:', (e))
        return None

    swarp.wait()

    # Add airmass to header:
    fits.setval(output_filename,'AIRMASS',value=airmass)

# def go_imagequality(inputfile,
#                 configfile=None,
#                 paramfile = None,
#                 convfile = 'default.conv',
#                 nnwfile = 'default.nnw',
#                 verbose=True, clean=True):
#     """
#     """
#

def go_register(filter_directories,
                lbc_chips = True,
                do_sextractor=True,
                do_scamp=True,
                do_swarp=True,
                astrometric_catalog='GAIA-DR2',
                scamp_iterations = 3):

    # TODO: Add the sextractor, scamp, swarp parameters for input.

    ###### Define which chips to extract if default is chosen:
    if lbc_chips == True:
        lbc_chips = [1,2,3,4]

    # If user enters just a single directory:
    if np.size(filter_directories) == 1 & ~isinstance(filter_directories,list):
        filter_directories = [filter_directories]

    for j in np.arange(np.size(filter_directories)):
        # Make sure the input directories have trailing slashes:
        drctry = filter_directories[j]
        if filter_directories[j][-1] != '/':
            filter_directories[j] += '/'


    # Loop through each of the filter directories:
    for fltdr in filter_directories:
        input_filenames = []

        # Only include the chips we want in the final image:
        for chp in lbc_chips:
            fls = glob(fltdr + '*_'+str(chp)+'.fits')
            for fl in fls:
                input_filenames.append(fl)

        # Loop through the files
        # go_sextractor = find sources
        # go_scamp = calculate astrometry
        for filename in input_filenames:
            # Find sources for alignment
            if do_sextractor:
                go_sextractor(filename)
            # Calculate the astrometry
            if do_scamp:
                go_scamp(filename, astrometric_catalog=astrometric_catalog,
                         num_iterations = scamp_iterations)

        # Stitch together the images
        # go_swarp = reproject and coadd images
        if do_swarp:
            go_swarp(input_filenames)



"""LBCgo: Data reduction package for the Large Binocular Telescope's Large Binocular Camera (LBC).

Requires astromatic.net software sextractor, scamp, swarp."""
__version__ = '0.1.3'

from .lbcproc import *
from .lbcregister import *



!cp /Users/howk/python/LBCgo/LBCgo/conf/* ./
%load_ext autoreload
%autoreload 2

#After adding the LBCgo directories to PYTHONPATH:
from lbcproc import *
from lbcregister import *

from astropy.io import fits
from ccdproc import CCDData, ImageFileCollection


lbc_chips=[1,2,3,4]
fltr_dirs=glob('I-BESSEL/')
go_register(fltr_dirs, lbc_chips=lbc_chips,
            do_sextractor=True, do_scamp=False, do_swarp=False)

go_register(fltr_dirs, lbc_chips=lbc_chips,scamp_iterations=3,
            do_sextractor=False, do_scamp=True, do_swarp=False)




## Found an image with *2* high S/N objects in the SCAMP solution.
!cp I-BESSEL/lbcr.20100318.091047_3.fits Test/

lbc_chips=[1,2,3,4]
fltr_dirs=glob('Test/')
go_register(fltr_dirs, lbc_chips=lbc_chips,
            do_sextractor=True, do_scamp=False, do_swarp=False)

go_register(fltr_dirs, lbc_chips=lbc_chips,scamp_iterations=4,
            astroref_catalog='GSC-2.3',
            do_sextractor=False, do_scamp=True, do_swarp=False)

go_register(fltr_dirs, lbc_chips=lbc_chips,scamp_iterations=5,
            do_sextractor=False, do_scamp=True, do_swarp=False)


go_register(fltr_dirs, lbc_chips=lbc_chips,scamp_iterations=3,
            do_sextractor=False, do_scamp=False, do_swarp=True)






#Try to do this with multiple exposures using the INSTRUMENT stability type.

scamp \
I-BESSEL/lbcr.20100318.091047_1.cat \
I-BESSEL/lbcr.20100318.092032_1.cat \
I-BESSEL/lbcr.20100318.093008_1.cat \
I-BESSEL/lbcr.20100318.095531_1.cat \
I-BESSEL/lbcr.20100319.094300_1.cat \
I-BESSEL/lbcr.20100319.095307_1.cat \
I-BESSEL/lbcr.20100318.091047_2.cat \
I-BESSEL/lbcr.20100318.092032_2.cat \
I-BESSEL/lbcr.20100318.093008_2.cat \
I-BESSEL/lbcr.20100318.095531_2.cat \
I-BESSEL/lbcr.20100319.094300_2.cat \
I-BESSEL/lbcr.20100319.095307_2.cat \
I-BESSEL/lbcr.20100318.091047_3.cat \
I-BESSEL/lbcr.20100318.092032_3.cat \
I-BESSEL/lbcr.20100318.093008_3.cat \
I-BESSEL/lbcr.20100318.095531_3.cat \
I-BESSEL/lbcr.20100319.094300_3.cat \
I-BESSEL/lbcr.20100319.095307_3.cat \
I-BESSEL/lbcr.20100318.091047_4.cat \
I-BESSEL/lbcr.20100318.092032_4.cat \
I-BESSEL/lbcr.20100318.093008_4.cat \
I-BESSEL/lbcr.20100318.095531_4.cat \
I-BESSEL/lbcr.20100319.094300_4.cat \
I-BESSEL/lbcr.20100319.095307_4.cat \
-c scamp.lbc.conf  -STABILITY_TYPE INSTRUMENT \
-PIXSCALE_MAXERR 1.2 -POSANGLE_MAXERR 3.0 -POSITION_MAXERR 1.0 \
-DISTORT_DEGREES 3 \
-MOSAIC_TYPE LOOSE \
-ASTREF_CATALOG GAIA-DR1 -AHEADER_SUFFIX .ahead \
-CROSSID_RADIUS 10.0

scamp \
I-BESSEL/lbcr.20100318.091047_1.cat \
I-BESSEL/lbcr.20100318.092032_1.cat \
I-BESSEL/lbcr.20100318.093008_1.cat \
I-BESSEL/lbcr.20100318.095531_1.cat \
I-BESSEL/lbcr.20100319.094300_1.cat \
I-BESSEL/lbcr.20100319.095307_1.cat \
I-BESSEL/lbcr.20100318.091047_2.cat \
I-BESSEL/lbcr.20100318.092032_2.cat \
I-BESSEL/lbcr.20100318.093008_2.cat \
I-BESSEL/lbcr.20100318.095531_2.cat \
I-BESSEL/lbcr.20100319.094300_2.cat \
I-BESSEL/lbcr.20100319.095307_2.cat \
I-BESSEL/lbcr.20100318.091047_3.cat \
I-BESSEL/lbcr.20100318.092032_3.cat \
I-BESSEL/lbcr.20100318.093008_3.cat \
I-BESSEL/lbcr.20100318.095531_3.cat \
I-BESSEL/lbcr.20100319.094300_3.cat \
I-BESSEL/lbcr.20100319.095307_3.cat \
I-BESSEL/lbcr.20100318.091047_4.cat \
I-BESSEL/lbcr.20100318.092032_4.cat \
I-BESSEL/lbcr.20100318.093008_4.cat \
I-BESSEL/lbcr.20100318.095531_4.cat \
I-BESSEL/lbcr.20100319.094300_4.cat \
I-BESSEL/lbcr.20100319.095307_4.cat \
-c scamp.lbc.conf  -STABILITY_TYPE INSTRUMENT \
-PIXSCALE_MAXERR 1.1 -POSANGLE_MAXERR 1.0 -POSITION_MAXERR 0.5 \
-DISTORT_DEGREES 3 \
-MOSAIC_TYPE FIX_FOCALPLANE \
-ASTREF_CATALOG GAIA-DR1 -AHEADER_SUFFIX .head \
-CROSSID_RADIUS 5.0


######################################################################

scamp \
I-BESSEL/lbcr.20100318.091047_2.cat \
I-BESSEL/lbcr.20100318.092032_2.cat \
I-BESSEL/lbcr.20100318.093008_2.cat \
I-BESSEL/lbcr.20100318.095531_2.cat \
I-BESSEL/lbcr.20100319.094300_2.cat \
I-BESSEL/lbcr.20100319.095307_2.cat \
-c scamp.lbc.conf  -STABILITY_TYPE INSTRUMENT \
-PIXSCALE_MAXERR 1.1 -POSANGLE_MAXERR 1.0 -POSITION_MAXERR 1.0 \
-DISTORT_DEGREES 3 \
-MOSAIC_TYPE LOOSE \
-ASTREF_CATALOG GAIA-DR1 -AHEADER_SUFFIX .ahead \
-CROSSID_RADIUS 10.0 

scamp \
I-BESSEL/lbcr.20100318.091047_2.cat \
I-BESSEL/lbcr.20100318.092032_2.cat \
I-BESSEL/lbcr.20100318.093008_2.cat \
I-BESSEL/lbcr.20100318.095531_2.cat \
I-BESSEL/lbcr.20100319.094300_2.cat \
I-BESSEL/lbcr.20100319.095307_2.cat \
-c scamp.lbc.conf  -STABILITY_TYPE INSTRUMENT \
-PIXSCALE_MAXERR 1.1 -POSANGLE_MAXERR 1.0 -POSITION_MAXERR 0.5 \
-DISTORT_DEGREES 3 \
-MOSAIC_TYPE FIX_FOCALPLANE \
-ASTREF_CATALOG GAIA-DR1 -AHEADER_SUFFIX .head \
-CROSSID_RADIUS 5.0 

scamp \
I-BESSEL/lbcr.20100318.091047_2.cat \
I-BESSEL/lbcr.20100318.092032_2.cat \
I-BESSEL/lbcr.20100318.093008_2.cat \
I-BESSEL/lbcr.20100318.095531_2.cat \
I-BESSEL/lbcr.20100319.094300_2.cat \
I-BESSEL/lbcr.20100319.095307_2.cat \
-c scamp.lbc.conf  -STABILITY_TYPE INSTRUMENT \
-PIXSCALE_MAXERR 1.025 -POSANGLE_MAXERR 1.0 -POSITION_MAXERR 0.5 \
-DISTORT_DEGREES 3 \
-MOSAIC_TYPE FIX_FOCALPLANE \
-ASTREF_CATALOG GAIA-DR1 -AHEADER_SUFFIX .head \
-CROSSID_RADIUS 1.0 -MATCH N

##### 3
scamp \
I-BESSEL/lbcr.20100318.091047_3.cat \
I-BESSEL/lbcr.20100318.092032_3.cat \
I-BESSEL/lbcr.20100318.093008_3.cat \
I-BESSEL/lbcr.20100318.095531_3.cat \
I-BESSEL/lbcr.20100319.094300_3.cat \
I-BESSEL/lbcr.20100319.095307_3.cat \
-c scamp.lbc.conf  -STABILITY_TYPE INSTRUMENT \
-PIXSCALE_MAXERR 1.1 -POSANGLE_MAXERR 1.0 -POSITION_MAXERR 1.0 \
-DISTORT_DEGREES 3 \
-MOSAIC_TYPE LOOSE \
-ASTREF_CATALOG GAIA-DR1 -AHEADER_SUFFIX .ahead \
-CROSSID_RADIUS 10.0 

scamp \
I-BESSEL/lbcr.20100318.091047_3.cat \
I-BESSEL/lbcr.20100318.092032_3.cat \
I-BESSEL/lbcr.20100318.093008_3.cat \
I-BESSEL/lbcr.20100318.095531_3.cat \
I-BESSEL/lbcr.20100319.094300_3.cat \
I-BESSEL/lbcr.20100319.095307_3.cat \
-c scamp.lbc.conf  -STABILITY_TYPE INSTRUMENT \
-PIXSCALE_MAXERR 1.1 -POSANGLE_MAXERR 1.0 -POSITION_MAXERR 0.5 \
-DISTORT_DEGREES 3 \
-MOSAIC_TYPE FIX_FOCALPLANE \
-ASTREF_CATALOG GAIA-DR1 -AHEADER_SUFFIX .head \
-CROSSID_RADIUS 5.0 -MATCH N

scamp \
I-BESSEL/lbcr.20100318.091047_3.cat \
I-BESSEL/lbcr.20100318.092032_3.cat \
I-BESSEL/lbcr.20100318.093008_3.cat \
I-BESSEL/lbcr.20100318.095531_3.cat \
I-BESSEL/lbcr.20100319.094300_3.cat \
I-BESSEL/lbcr.20100319.095307_3.cat \
-c scamp.lbc.conf  -STABILITY_TYPE INSTRUMENT \
-PIXSCALE_MAXERR 1.025 -POSANGLE_MAXERR 1.0 -POSITION_MAXERR 0.5 \
-DISTORT_DEGREES 3 \
-MOSAIC_TYPE FIX_FOCALPLANE \
-ASTREF_CATALOG GAIA-DR1 -AHEADER_SUFFIX .head \
-CROSSID_RADIUS 1.0 -MATCH N




############################################################################
#                  Image reduction with CCDPROC package                    #
#                                                                          #
#                             VERSION: 0.0.5                               #
#                                                                          #
#                          Ruben Garcia-Benito                             #
#                                RGB@IAA                                   #
#                                                                          #
#                       Last Change: 2017/02/27                            #
#                                                                          #
#                                                                          #
# See https://github.com/rgbIAA/reduceccd                                  #
############################################################################

from photutils import Background2D, SigmaClip, MedianBackground, MeanBackground
from photutils import SExtractorBackground, BiweightLocationBackground
from photutils import ModeEstimatorBackground, MMMBackground
from ccdproc.utils.slices import slice_from_string
from skimage.feature import register_translation
from astropy.stats import sigma_clipped_stats
from scipy.ndimage import interpolation
from ccdproc import ImageFileCollection
from photutils import make_source_mask
import astropy.io.fits as pyfits
from astropy.table import Table
from astropy import units as u
from ccdproc import CCDData
from copy import deepcopy
# from pyraf import iraf
import numpy as np
import ccdproc
import sys
import os
import re

################################ VERSION ###################################
VERSION = '0.0.5'  #


############################################################################


# ------------------------------ FUNCTIONS -----------------------------
def join_path(name, directory=None):
    if directory is not None:
        name = os.path.join(directory, os.path.basename(name))
    return name


def fits2CCDData(lfits, key_unit='BUNIT', key_file='FILENAME', unit=None, single=False):
    lccd = []
    if not isinstance(lfits, (tuple, list)):
        lfits = [lfits]
    for fits in lfits:
        fits_unit = unit
        if os.path.exists(fits):
            hdr = pyfits.getheader(fits)
        else:
            print ('>>> WARNING: File "%s" NOT found' % os.path.basename(fits))
            continue
        if key_unit is not None and key_unit in hdr:
            try:
                fits_unit = eval('u.%s' % hdr[key_unit])
            except:
                pass
        if fits_unit is None:
            if key_unit is not None:
                sys.exit('>>> Units NOT found in header ("%s") of image "%s". Specify one in "unit" variable' % (
                key_unit, os.path.basename(fits)))
            else:
                print ('>>> WARNING: "key_unit" not specified')
        ccd = CCDData.read(fits, unit=fits_unit)
        if key_file is not None and not key_file in ccd.header:
            ccd.header[key_file] = os.path.basename(fits)
        lccd.append(ccd)
    if len(lccd) == 0:
        print ('>>> WARNING: NO files found!')
        return
    if single and len(lccd) == 1:
        lccd = lccd[0]
    return lccd


def getObjects(image_file_collection):
    table = image_file_collection.summary
    pattern = re.compile('flat|bias|bd\+|feige|land')
    objects = [item.split('-')[0].split('.')[-1] for item in table['file'] if pattern.search(item.lower()) is None]
    return np.unique(objects).tolist()


def getFilters(image_file_collection, key_filter='filter'):
    table = image_file_collection.summary
    filters = np.unique(table[key_filter].data)
    if isinstance(filters, np.ma.MaskedArray):
        filters = np.sort(filters[~filters.mask])
    return filters.tolist()


def getFilename(ccdfile, key='FILENAME'):
    namefile = None
    if isinstance(ccdfile, str):
        namefile = ccdfile
    elif isinstance(ccdfile, CCDData):
        if key in ccdfile.header:
            namefile = ccdfile.header[key]
    return namefile


def addKeyHdr(hdr, key, value):
    if value is not None and key is not None:
        hdr[key] = value
    return hdr


def ImageFileCollectionFilter(image_file_collection, ldfilter={}, dirname=None, abspath=True,
                              key_find='find', invert_find=False, return_mask=False, key='file',
                              dkeys=None, copy=True):
    dirname = image_file_collection.location if dirname is None else dirname
    if not abspath:
        dirname = None
    if not isinstance(ldfilter, (tuple, list)):
        ldfilter = [ldfilter]
    if dkeys is not None:
        ldfilter = addKeysListDict(ldfilter, dkeys, copy=True)
    mask = None
    lmask = []
    for dfilter in ldfilter:
        if dfilter is None or len(dfilter) == 0:
            continue
        lfmask = []
        if key_find in dfilter:
            key_string = dfilter.pop(key_find)
            lfiles_find_mask = np.char.find(image_file_collection.summary[key].data, key_string) > -1
            if invert_find:
                lfiles_find_mask = np.invert(lfiles_find_mask)
            lfmask.append(lfiles_find_mask)
        lfiles = image_file_collection.files_filtered(**dfilter)
        lfiles_mask = np.in1d(image_file_collection.summary['file'], lfiles)
        lfmask.append(lfiles_mask)
        fmask = np.logical_and.reduce(lfmask)
        lmask.append(fmask)
    if len(lmask) > 0:
        mask = np.logical_or.reduce(lmask)
        list_files = image_file_collection.summary['file'][mask]
    else:
        list_files = image_file_collection.summary['file']
    list_files = [join_path(image, dirname) for image in list_files]
    if return_mask:
        return list_files, mask
    else:
        return list_files


def getListFiles(list_files, dfilter=None, mask=None, key_find='find', invert_find=False):
    if isinstance(list_files, ImageFileCollection):
        if mask is None:
            list_files = ImageFileCollectionFilter(list_files, dfilter,
                                                   key_find=key_find, invert_find=invert_find)
        else:
            list_files = [join_path(image, list_files.location) for image in list_files.summary['file'][mask]]
    return list_files


def getTimeTable(list_files, key_time='EXPTIME', key_file='file', ext=0, abspath=False, mask=-1, clean=True, sort=True,
                 **kwargs):
    dt = {key_file: [], key_time: []}
    for fits in list_files:
        fits = fits if abspath else os.path.basename(fits)
        dt[key_file].append(fits)
        hdr = pyfits.getheader(fits, ext=ext, **kwargs)
        if key_time in hdr:
            dt[key_time].append(hdr[key_time])
        else:
            dt[key_time].append(mask)
    table = Table(dt)
    if clean:
        id_time = table[key_time] == mask
        table = table[np.invert(id_time)]
    if sort:
        table.sort(key_time)
    return table


def addKeysListDict(ldict, dkeys, copy=True, force=False):
    if ldict is not None and dkeys is not None:
        if not isinstance(ldict, (list, tuple)):
            ldict = [ldict]
        if copy:
            ldict = deepcopy(ldict)
        for i in range(len(ldict)):
            if ldict[i] is not None and isinstance(ldict[i], dict):
                for key in dkeys:
                    if key is not None:
                        if not key in ldict[i] or (key in ldict[i] and force):
                            ldict[i][key] = dkeys[key]
    return ldict


def ammendHeader(header):
    if '' in header:
        del header['']
    return header


def cleanCosmic(ccd, mbox=15, rbox=15, gbox=11, sigclip=5, cleantype="medmask", cosmic_method='lacosmic'):
    ctype = cosmic_method.lower().strip()
    ctypes = ['lacosmic', 'median']
    if not ctype in ctypes:
        print ('>>> Cosmic ray type "%s" NOT available [%s]' % (ctype, ' | '.join(ctypes)))
        return
    if ctype == 'lacosmic':
        ccd = ccdproc.cosmicray_lacosmic(ccd, sigclip=sigclip, cleantype=cleantype)
    elif ctype == 'median':
        ccd = ccdproc.cosmicray_median(ccd, mbox=mbox, rbox=rbox, gbox=gbox)
    if isinstance(ccd, CCDData):
        ccd.header['COSMIC'] = ctype.upper()
    return ccd


def cropImage(image, trim=None, zoom=None, center=None, pixels=None, width=None):
    if isinstance(image, str):
        image = fits2CCDData(image, single=True).data
    if trim is not None:
        return image[slice_from_string(trim)]
    if center is None:
        center = (image.shape[0] / 2, image.shape[1] / 2)
    dy, dx = image.shape
    if zoom is not None:
        if isinstance(zoom, (list, tuple)):
            xzoom, yzoom = zoom
        else:
            xzoom, yzoom = zoom, zoom
        if xzoom < 1.0 or yzoom < 1.0:
            print ('WARNING: Zoom needs to be >= 1.0 [%s, %s]! Nothing done!' % (xzoom, yzoom))
            return image
        dy /= yzoom
        dx /= xzoom
        idy = slice(center[0] - int(dy / 2), center[0] + int(dy / 2), None)
        idx = slice(center[1] - int(dx / 2), center[1] + int(dx / 2), None)
        image = image[idy, idx]
    elif pixels is not None:
        if isinstance(pixels, (tuple, list, np.ndarray)):
            ypix, xpix = pixels
        else:
            ypix, xpix = pixels, pixels
        if ypix is not None:
            image = image[ypix: dy - ypix, ...]
        if xpix is not None:
            image = image[..., xpix: dx - xpix]
    elif width is not None:
        if isinstance(width, (tuple, list, np.ndarray)):
            ywidth, xwidth = width
        else:
            ywidth, xwidth = width, width
        if ywidth is not None:
            idy = slice(center[0] - int(ywidth / 2), center[0] + int(ywidth / 2), None)
            image = image[idy, ...]
        if xwidth is not None:
            idx = slice(center[1] - int(xwidth / 2), center[1] + int(xwidth / 2), None)
            image = image[..., idx]
    return image


def shiftImage(image, ref_image, precision=100, shift=None, trim=None, zoom=None, center=None,
               pixels=None, width=None, ref_trim=None, ref_zoom=None, ref_center=None, ref_pixels=None,
               ref_width=None, copy_crop=True, copy=True):
    if isinstance(image, str):
        image = fits2CCDData(image, single=True)
    if isinstance(ref_image, str):
        ref_image = fits2CCDData(ref_image, single=True)
    if isinstance(ref_image, CCDData):
        ref_image = ref_image.data
    if copy:
        image = image.copy()
    if shift is None:
        if copy_crop:
            if ref_trim is None:
                ref_trim = trim
            if ref_zoom is None:
                ref_zoom = zoom
            if ref_center is None:
                ref_center = center
            if ref_pixels is None:
                ref_pixels = pixels
            if ref_width is None:
                ref_width = width
        idata = cropImage(image.data, trim=trim, zoom=zoom, center=center, pixels=pixels, width=width)
        rdata = cropImage(ref_image, trim=ref_trim, zoom=ref_zoom, center=ref_center, pixels=ref_pixels,
                          width=ref_width)
        shift, error, diffphase = register_translation(idata, rdata, precision)
    else:
        error = None
    image.data = interpolation.shift(image.data, (-shift[0], -shift[1]))
    if image.uncertainty is not None:
        image.uncertainty.array = interpolation.shift(image.uncertainty.array, (-shift[0], -shift[1]))
    if image.mask is not None:
        image.mask = interpolation.shift(image.mask, (-shift[0], -shift[1]))
    image.header['YSHIFT'] = -shift[0]
    image.header['XSHIFT'] = -shift[1]
    if error is not None:
        image.header['PRECISION'] = precision
        image.header['ESHIFT'] = error
        image.header['DIFFPHASE'] = diffphase
    return image


def create_master_bias(list_files, fitsfile=None, fits_section=None, gain=None, method='median',
                       dfilter={'imagetyp': 'bias'}, mask=None, key_find='find', invert_find=False, sjoin=','):
    if gain is not None and not isinstance(gain, u.Quantity):
        gain = gain * u.electron / u.adu
    lbias = []
    list_files = getListFiles(list_files, dfilter, mask, key_find=key_find, invert_find=invert_find)
    for filename in list_files:
        ccd = CCDData.read(filename, unit=u.adu)
        trimmed = True if fits_section is not None else False
        ccd = ccdproc.trim_image(ccd, fits_section=fits_section, add_keyword={'trimmed': trimmed})
        if gain is not None:
            ccd = ccdproc.gain_correct(ccd, gain)
        lbias.append(ccd)
    combine = ccdproc.combine(lbias, method=method)
    if gain is not None and not 'GAIN' in combine.header:
        combine.header.set('GAIN', gain.value, gain.unit)
    combine.header['CGAIN'] = True if gain is not None else False
    combine.header['IMAGETYP'] = 'BIAS'
    combine.header['CMETHOD'] = method
    combine.header['CCDVER'] = VERSION
    if sjoin is not None:
        combine.header['LBIAS'] = sjoin.join([os.path.basename(fits) for fits in list_files])
    combine.header['NBIAS'] = len(list_files)
    if fitsfile is not None:
        combine.header['FILENAME'] = os.path.basename(fitsfile)
        combine.write(fitsfile, clobber=True)
    return combine


def create_master_flat(list_files, flat_filter=None, fitsfile=None, bias=None, fits_section=None, gain=None,
                       method='median', key_filter='filter', dfilter={'imagetyp': 'FLAT'}, mask=None, key_find='find',
                       invert_find=False, sjoin=','):
    if gain is not None and not isinstance(gain, u.Quantity):
        gain = gain * u.electron / u.adu
    lflat = []
    if dfilter is not None and key_filter is not None and flat_filter is not None:
        dfilter = addKeysListDict(dfilter, {key_filter: flat_filter})
    list_files = getListFiles(list_files, dfilter, mask, key_find=key_find, invert_find=invert_find)
    if len(list_files) == 0:
        print ('WARNING: No FLAT files available for filter "%s"' % flat_filter)
        return
    for filename in list_files:
        ccd = CCDData.read(filename, unit=u.adu)
        trimmed = True if fits_section is not None else False
        ccd = ccdproc.trim_image(ccd, fits_section=fits_section, add_keyword={'trimmed': trimmed})
        if gain is not None:
            ccd = ccdproc.gain_correct(ccd, gain)
        if bias is not None:
            if isinstance(bias, str):
                bias = fits2CCDData(bias, single=True)
            ccd = ccdproc.subtract_bias(ccd, bias)
        lflat.append(ccd)
    combine = ccdproc.combine(lflat, method=method)
    if gain is not None and not 'GAIN' in combine.header:
        combine.header.set('GAIN', gain.value, gain.unit)
    combine.header['CGAIN'] = True if gain is not None else False
    combine.header['IMAGETYP'] = 'FLAT'
    combine.header['CMETHOD'] = method
    combine.header['CCDVER'] = VERSION
    addKeyHdr(combine.header, 'MBIAS', getFilename(bias))
    if sjoin is not None:
        combine.header['LFLAT'] = sjoin.join([os.path.basename(fits) for fits in list_files])
    combine.header['NFLAT'] = len(list_files)
    if fitsfile is not None:
        combine.header['FILENAME'] = os.path.basename(fitsfile)
        combine.write(fitsfile, clobber=True)
    return combine


def create_master_flat_from_dict(list_files, dflat, verbose=True, **kwargs):
    dflat_ccd = {}
    for flat_filter in dflat:
        master_flat = dflat[flat_filter]
        if verbose:
            print ('>>> Creating FLAT: %s' % os.path.basename(master_flat))
        dflat_ccd[flat_filter] = create_master_flat(list_files, flat_filter, master_flat, **kwargs)
    return dflat_ccd


def ccdproc_images_filter(list_files, image_filter=None, master_flat=None, master_bias=None, fits_section=None,
                          gain=None, readnoise=None,
                          error=False, sky=True, dout=None, cosmic=False, mbox=15, rbox=15, gbox=11,
                          cleantype="medmask", cosmic_method='lacosmic',
                          sigclip=5, key_filter='filter', dfilter={'imagetyp': 'LIGHT'},
                          mask=None, key_find='find',
                          invert_find=False, **kwargs):
    if error and (gain is None or readnoise is None):
        print ('WARNING: You need to provide "gain" and "readnoise" to compute the error!')
        return
    if gain is not None and not isinstance(gain, u.Quantity):
        gain = gain * u.electron / u.adu
    if readnoise is not None and not isinstance(readnoise, u.Quantity):
        readnoise = readnoise * u.electron
    if dfilter is not None and key_filter is not None and image_filter is not None:
        dfilter = addKeysListDict(dfilter, {key_filter: image_filter})
    list_files = getListFiles(list_files, dfilter, mask, key_find=key_find, invert_find=invert_find)
    dccd = {}
    for filename in list_files:
        ccd = CCDData.read(filename, unit=u.adu)
        nccd = ccdproc.ccd_process(ccd, trim=fits_section, gain=gain, master_bias=master_bias, master_flat=master_flat,
                                   readnoise=readnoise, error=error)
        for key in ccd.header:
            if not key in nccd.header:
                nccd.header[key] = ccd.header[key]
        # Better get rid of the cosmic rays BEFORE subtracting the global sky background
        if cosmic:
            nccd = cleanCosmic(nccd, mbox=mbox, rbox=rbox, gbox=gbox, sigclip=sigclip, cleantype=cleantype,
                               cosmic_method=cosmic_method)
        if sky:
            nccd = subtract_sky_ccd(nccd, **kwargs)
        addKeyHdr(nccd.header, 'MBIAS', getFilename(master_bias))
        addKeyHdr(nccd.header, 'MFLAT', getFilename(master_flat))
        filename = 'c%s' % os.path.basename(filename)
        dccd[filename] = nccd
        filename = join_path(filename, dout)
        nccd.header['FILENAME'] = os.path.basename(filename)
        nccd.header['CCDVER'] = VERSION
        nccd.header = ammendHeader(nccd.header)
        nccd.write(filename, clobber=True)
    return dccd


def ccdproc_images(list_files, dmaster_flat, master_bias=None,
                   fits_section=None, sky=True, verbose=True, **kwargs):
    dccd_images = {}
    for filt in dmaster_flat:
        if verbose:
            print ('>>> Reducing Filter: %s' % filt)
        dccd = ccdproc_images_filter(list_files, filt, master_flat=dmaster_flat[filt], master_bias=master_bias,
                                     fits_section=fits_section, sky=sky, **kwargs)
        dccd_images[filt] = dccd
    return dccd_images


def subtract_sky_ccd_percentile(ccd, skysub='SKYSUB', skyval='SKYVAL', lower=1., upper=95., check_skysub=True,
                                max_val=10000):
    if check_skysub and skysub is not None and skysub in ccd.header and ccd.header[skysub]:
        return ccd
    if not check_skysub and skysub is not None and skysub in ccd.header and ccd.header[skysub]:
        print (
        'WARNING: Image already sky subtracted! Subtracting AGAIN the sky! [set check_skysub = True for checking]')
    data = ccd.data[np.isfinite(ccd.data) & (ccd.data > 0.)].flatten()
    percentile = np.percentile(data, [lower, upper])
    sky = np.median(percentile)
    if max_val is not None and sky >= max_val:
        if 'FILENAME' in ccd.header:
            print ('WARNING: File "%s" has high sky level (%s >= %s). Correction NOT applied' % (
            ccd.header['FILENAME'], sky, max_val))
        else:
            print ('WARNING: File has high sky level (%s >= %s). Correction NOT applied' % (sky, max_val))
        return ccd
    ccd.data -= sky
    ccd.header[skysub] = True
    ccd.header[skyval] = sky
    ccd.header['SKYTYPE'] = '1D'
    return ccd


def subtract_sky_ccd_mask(ccd, skysub='SKYSUB', skyval='SKYVAL', snr=3, npixels=5, dilate_size=11, sigma=3,
                          check_skysub=True, max_val=10000):
    if check_skysub and skysub is not None and skysub in ccd.header and ccd.header[skysub]:
        return ccd
    if not check_skysub and skysub is not None and skysub in ccd.header and ccd.header[skysub]:
        print (
        'WARNING: Image already sky subtracted! Subtracting AGAIN the sky! [set check_skysub = True for checking]')
    mask = make_source_mask(ccd.data, snr=sigma, npixels=npixels, dilate_size=dilate_size)
    mean, median, std = sigma_clipped_stats(ccd.data, sigma=sigma, mask=mask)
    sky = median
    if max_val is not None and sky >= max_val:
        if 'FILENAME' in ccd.header:
            print ('WARNING: File "%s" has high sky level (%s >= %s). Correction NOT applied' % (
            ccd.header['FILENAME'], sky, max_val))
        else:
            print ('WARNING: File has high sky level (%s >= %s). Correction NOT applied' % (sky, max_val))
        return ccd
    ccd.data -= sky
    ccd.header[skysub] = True
    ccd.header[skyval] = sky
    ccd.header['SKYTYPE'] = '1D'
    return ccd


def subtract_sky_ccd_2d(ccd, skysub='SKYSUB', skyval='SKYVAL', box_size=(50, 50), filter_size=(3, 3),
                        bkg_method='sextractor',
                        check_skysub=True, max_val=10000, fitsky=None, sigma=3., iters=10, edge_method='crop'):
    # Although edge_method='pad' is recommended, it does give errors "total size of new array must be unchanged" for v0.3
    if check_skysub and skysub is not None and skysub in ccd.header and ccd.header[skysub]:
        return ccd
    if not check_skysub and skysub is not None and skysub in ccd.header and ccd.header[skysub]:
        print (
        'WARNING: Image already sky subtracted! Subtracting AGAIN the sky! [set check_skysub = True for checking]')

    dsky = {'median': MedianBackground(), 'mean': MeanBackground(), 'sextractor': SExtractorBackground(),
            'biweight': BiweightLocationBackground(), 'mode': ModeEstimatorBackground(), 'mm': MMMBackground()}

    if not bkg_method in dsky:
        print ('WARNING: Background type "%s" NOT available!! Selecting "median" from [%s]' % (
        bkg_type, ' | '.join(dksy.keys())))
        bkg_estimator = dksy['median']
    else:
        bkg_estimator = dsky[bkg_method]

    sigma_clip = SigmaClip(sigma=sigma, iters=iters) if sigma is not None and iters is not None else None

    bkg = Background2D(ccd.data, box_size, filter_size=filter_size, sigma_clip=sigma_clip, bkg_estimator=bkg_estimator,
                       edge_method=edge_method)

    if max_val is not None and bkg.background_median >= max_val:
        if 'FILENAME' in ccd.header:
            print ('WARNING: File "%s" has high sky level (%s >= %s). Correction NOT applied' % (
            ccd.header['FILENAME'], bkg.background_median, max_val))
        else:
            print (
            'WARNING: File has high sky level (%s >= %s). Correction NOT applied' % (bkg.background_median, max_val))
        return ccd
    ccd.data -= bkg.background
    ccd.header[skysub] = True
    ccd.header[skyval] = bkg.background_median
    ccd.header['SKYTYPE'] = '2D'
    ccd.header['BKGTYPE'] = bkg_method
    ccd.bkg = bkg
    if fitsky is not None:
        pyfits.writeto(fitsky, bkg.background, clobber=True)
    return ccd


def subtract_sky_ccd(ccd, method='2d', **kwargs):
    method = method.strip().lower()
    methods = ['percentile', 'mask', '2d']
    if method == 'percentile':
        return subtract_sky_ccd_percentile(ccd, **kwargs)
    elif method == 'mask':
        return subtract_sky_ccd_mask(ccd, **kwargs)
    elif method == '2d':
        return subtract_sky_ccd_2d(ccd, **kwargs)
    else:
        print ('>>> Method "%s" NOT found! Selecting "2d" from [%s]' % (method, ' | '.join(methods)))
        return subtract_sky_ccd_2d(ccd, **kwargs)


def subtract_sky_ccds(lccd, method='2d', **kwargs):
    if not isinstance(lccd, (tuple, list)):
        lcdd = [lccd]
    nlccd = []
    for ccd in lccd:
        nlccd.append(subtract_sky_ccd(ccd, method, **kwargs))
    return nlccd


def align_combine_images(list_files, fitsfile, ref_image_fits=None, precision=100, dout=None, hexp='EXPTIME',
                         force=False,
                         minmax_clip=False, minmax_clip_min=0.0, sigma_clip=True, date=None, clip_extrema=False,
                         func=np.ma.median, sigclip=5,
                         cosmic=False, mbox=15, rbox=15, gbox=11, cleantype="medmask", cosmic_method='lacosmic',
                         sky=False, dict_sky={},
                         dict_combine={}, suffix=None, hfilter='FILTER', key_file='file', hobj='OBJECT', ext=0,
                         method='median',
                         align=True, **kwargs):
    if ref_image_fits is None:
        tobj = getTimeTable(list_files, key_time=hexp, key_file=key_file, ext=ext, abspath=True, mask=-1, sort=True,
                            clean=True)
        ref_image_fits, list_files = tobj[key_file][-1], tobj[key_file][:-1].data.tolist()
    ref_image = CCDData.read(ref_image_fits)
    ref_image = ccdproc.cosmicray_lacosmic(ref_image, sigclip=sigclip)
    lccd = [ref_image]
    lexp = [ref_image.header[hexp]]
    all_images = [os.path.basename(ref_image_fits)]
    for img in list_files:
        image = fits2CCDData(img, single=True)
        image_suffix = suffix if suffix is not None else img.split('.')[-1]
        if 'REFIMA' in image.header and 'IMAGES' in image.header and not force:
            continue
        if cosmic:
            image = cleanCosmic(image, mbox=mbox, rbox=rbox, gbox=gbox, sigclip=sigclip, cleantype=cleantype,
                                cosmic_method=cosmic_method)
        if align:
            offset_image = shiftImage(image, ref_image, precision=precision, **kwargs)
            img_shifted = '%s_shift.%s' % (img.split('.fit')[0], image_suffix)
            all_images.append(os.path.basename(img_shifted))
            img_shifted = join_path(img_shifted, dout)
            offset_image.write(img_shifted, clobber=True)
            lccd.append(offset_image)
        else:
            all_images.append(os.path.basename(img))
            lccd.append(image)
        lexp.append(image.header[hexp])
        # iraf.unlearn("imshift")
        # iraf.imshift.interp_type = "spline3"
        # iraf.imshift(img, img, shift[1], shift[0])
    # Combine images
    lexp = np.array(lexp)
    scale_func = None  # 1. / lexp
    combine = ccdproc.combine(lccd, method=method, scale=scale_func, minmax_clip=minmax_clip, func=func,
                              minmax_clip_min=minmax_clip_min, sigma_clip=sigma_clip, clip_extrema=clip_extrema,
                              **dict_combine)
    if sky:
        combine = subtract_sky_ccd(combine, **dict_sky)
    combine.header['IMAGES'] = str(' | '.join(all_images))
    combine.header['REFIMA'] = os.path.basename(ref_image_fits)
    combine.header['IMGSEXP'] = ' | '.join(map(str, lexp[1:].tolist() + [lexp[0]]))
    combine.header['CMETHOD'] = method
    dir_out = os.path.dirname(ref_image_fits) if dout is None else dout
    fitsfile = join_path(fitsfile, dir_out)
    combine.header['FILENAME'] = os.path.basename(fitsfile)
    combine.header['CCDVER'] = VERSION
    combine.write(fitsfile, clobber=True)


def align_combine(image_file_collection, filters=None, objects=None, dout=None, suffix=None, date=None, key_find='file',
                  key_filter='filter', invert_find=False, dfilter={'imagetyp': 'LIGHT'}, verbose=True, force=False,
                  find_obj=True,
                  find_filter=True, **kwargs):
    if filters is None:
        filters = getFilters(image_file_collection)
    if objects is None:
        objects = getObjects(image_file_collection)
    if not isinstance(filters, (tuple, list)):
        filters = [filters]
    if not isinstance(objects, (tuple, list)):
        objects = [objects]
    table = image_file_collection.summary
    ddata = image_file_collection.location
    for obj in objects:
        dkeys = {key_find: obj} if find_obj else None
        mask_object = ImageFileCollectionFilter(image_file_collection, dfilter, dkeys=dkeys, invert_find=invert_find,
                                                return_mask=True)[-1]
        if not np.any(mask_object):
            print ('WARNING: Object "%s" does NOT exists!!' % obj)
            continue
        for image_filter in filters:
            dkeys = {}
            if find_obj:
                dkeys[key_find] = obj
            if find_filter:
                dkeys[key_filter] = image_filter
            mask_table = ImageFileCollectionFilter(image_file_collection, dfilter, dkeys=dkeys, invert_find=invert_find,
                                                   return_mask=True)[-1]
            if 'refima' in table.colnames and 'images' in table.colnames and not force:
                mask = np.logical_and(table['refima'].mask, table['images'].mask)
                mask_table = np.logical_and(mask_table, mask)
            tobj = table[mask_table]
            tobj.sort('exposure')
            if tobj['file'].data.size == 0:
                print ('WARNING: Object "%s" does NOT contain filter "%s"' % (obj, image_filter))
                continue
            if verbose:
                print ('>>> OBJECT: %s (Filter %s)' % (obj, image_filter))
            ref_image_fits = join_path(tobj['file'][-1], image_file_collection.location)
            list_files = [join_path(fits, image_file_collection.location) for fits in tobj['file'][:-1].data]
            image_suffix = suffix if suffix is not None else ref_image_fits.split('.')[-1]
            fitsfile = '%s_%s.%s' % (obj, image_filter, image_suffix) if date is None or len(
                date) == 0 else '%s_%s_%s.%s' % (obj, image_filter, date, image_suffix)
            align_combine_images(list_files, fitsfile, ref_image_fits=ref_image_fits, dout=dout, force=force, **kwargs)


def reduceNight(path, filters=None, fits_section=None, date=None, dout=None, create_bias=True, create_flat=True,
                correct_images=True, gain=None, readnoise=None, sky=True, lower=1.0, upper=95., combine=True,
                align=True,
                objects=None, cosmic=True, mbox=15, rbox=15, gbox=11, cleantype="medmask", sky_after=True, dict_sky={},
                dict_combine={}, method='median', mask_bias=None, dfilter_bias={'imagetyp': 'bias'},
                lfits_bias=None,
                invert_find_bias=False, dfilter_flat={'imagetyp': 'FLAT'}, mask_flat=None,
                cosmic_method='lacosmic',
                invert_find_flat=False, dfilter_images={'imagetyp': 'LIGHT'}, mask_images=None,
                key_find='find',
                key_filter='filter', find_obj=True, invert_find_images=False, find_filter=True,
                error=False,
                invert_find_align=False, suffix=None, dict_align_combine={}, verbose=True):
    if date is None:
        date = path.split('/')[-2] if path.endswith('/') else path.split('/')[-1]
        if not date.isdigit():
            date = None
    if date is None:
        print ('>>> WARNING: You better provide a date!')

    if dout is not None and not os.path.exists(dout):
        dout = None
        print ('>>> WARNING: output directory does NOT exists! Saving to local working directory!')

    # Find all images
    ic_all = ImageFileCollection(path, keywords=None)

    # Find suffix
    if suffix is None:
        suffix = ic_all.summary['file'][0].split('.')[-1]

    if filters is None:
        filters = getFilters(ic_all)
        print ('>>> Filters found: %s' % ' | '.join(filters))
    if not isinstance(filters, (list, tuple, np.ndarray)):
        filters = [filters]

    if objects is None:
        objects = getObjects(ic_all)
        print ('>>> Objects found: %s' % ' | '.join(objects))
    if not isinstance(filters, (list, tuple, np.ndarray)):
        objects = [objects]

    # Sky option
    if sky:
        sky_before = not sky_after
    else:
        sky_before = False
        sky_after = False

    # Create name master bias
    master_bias = 'master_bias_%s.%s' % (date, suffix) if date is not None else 'master_bias.%s' % suffix
    master_bias = join_path(master_bias, dout)

    # Create name master Flats
    dflat = {}
    for filt in filters:
        master_flat = 'master_flat_%s_%s.%s' % (filt, date, suffix) if date is not None else 'master_flat_%s.%s' % (
        filt, suffix)
        dflat[filt] = join_path(master_flat, dout)

    # --------------- Master bias file ---------------------
    ccd_master_bias = None
    if create_bias and master_bias is not None:
        lfits_bias = ic_all if lfits_bias is None else lfits_bias
        if verbose:
            print ('>>> Creating BIAS: %s' % os.path.basename(master_bias))
        ccd_master_bias = create_master_bias(lfits_bias, master_bias, fits_section=fits_section, gain=gain,
                                             method=method, dfilter=dfilter_bias,
                                             mask=mask_bias, key_find=key_find,
                                             invert_find=invert_find_bias)
    if not create_bias and master_bias is not None:
        ccd_master_bias = fits2CCDData(master_bias, single=True)

    # -------- Create Master File for each filter ---------
    dccd_master_flat = None
    if create_flat:
        dccd_master_flat = create_master_flat_from_dict(ic_all, dflat, bias=ccd_master_bias,
                                                        fits_section=fits_section,
                                                        gain=gain, method=method, dfilter=dfilter_flat,
                                                        mask=mask_flat,
                                                        key_find=key_find, invert_find=invert_find_flat,
                                                        verbose=verbose)
    else:
        dccd_master_flat = {}
        for key in dflat:
            dccd_master_flat[key] = fits2CCDData(dflat[key], single=True)

    # ----------- Correct all science and standard stards images -------------
    if correct_images:
        ccdproc_images(ic_all, dccd_master_flat, master_bias=ccd_master_bias,
                       fits_section=fits_section, gain=gain,
                       dout=dout, sky=sky_before,
                       cosmic=cosmic, mbox=mbox, rbox=rbox, gbox=gbox, cleantype=cleantype,
                       cosmic_method=cosmic_method,
                       key_filter=key_filter,
                       dfilter=dfilter_images, mask=mask_images, key_find=key_find,
                       invert_find=invert_find_images,
                       verbose=verbose,
                       readnoise=readnoise, error=error, **dict_sky)

    # ----------- Align and combine -------------------
    if combine:
        align_file_collection = image_file_collection if dout is None else ImageFileCollection(dout, keywords=None)
        align_combine(align_file_collection, filters, objects, dout=dout, sky=sky_after, dict_sky=dict_sky,
                      dict_combine=dict_combine, key_find=key_find, invert_find=invert_find_align, suffix=suffix,
                      dfilter=dfilter_images, find_obj=find_obj, find_filter=find_filter, align=align,
                      verbose=verbose,
                      **dict_align_combine)

# ----------------------------------------------------------------------



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lbcgo', 'lbcgo.examples']

package_data = \
{'': ['*'], 'lbcgo': ['conf/*']}

install_requires = \
['astropy==0', 'ccdproc>=2.2.0,<3.0.0', 'numpy>=1.21.0,<2.0.0']

setup_kwargs = {
    'name': 'lbcgo',
    'version': '0.1.3',
    'description': "Data reduction for the LBT's Large Binocular Camera",
    'long_description': '# LBCgo: LBC data reduction pipeline\n\nWARNING: This code is currently under continued development. While the basic functionality exists, it should be used with some care and attention.\n\n## Dependencies:\n\nPython dependencies:\n* `astropy`\n* `CCDProc`\n* `numpy`\n\nExternal dependencies:\n* `SExtractor`\n* `SCAMP`\n* `SWarp`\n\nThe external C++ codes `SCAMP`, `SWarp`, and `SExtractor` developed by  Emmanuel Bertin and collaborators are available through http://astromatic.iap.fr. `SCAMP` and `SExtractor` are also available through GitHub: https://github.com/astromatic.\n\n## Running LBCgo:\n\nFor "standard" situations, the `LBCgo` can be run in one step from the python command line. In this case, all of the data in the `raw/` directory are taken on the same night and have appropriate calibrations. In this case, running `LBCgo` from the command line is as simple as:\n```\nipython> from lbcproc import *\nipython> from lbcregister import *\n\nipython> lbcgo()\n```\nBefore doing this, copy the parameter files from `LBCgo/LBCgo/conf/` into the current working directory (an eventual fix won\'t require this step).\n\nAlternatively, it can be useful to process each filter separately or even to avoid doing the astrometric steps until a later time. In this case, one may do:\n```\nipython> lbcgo(filter_names=[\'I-BESSEL\'], do_astrometry=False)\n```\n\nThe astrometric portion of the reduction can be done later using, for example reducing the I-BESSEL data for the target PG1338+101:\n```\nipython> fltr_dirs=glob(\'PG1338+101/I-BESSEL/\')\nipython> go_register(fltr_dirs, do_sextractor=True,\n            do_scamp=True, do_SWarp=True)\n```\n\n#### Missing chips:\n\n`LBCgo` can be used if the images were taken when one of the LBC CCDs was off-line. The approach to doing this is to explicitly specify the chips to include in the data reduction steps:\n```\nipython> lbcgo(lbc_chips=[1,2,4])\n```\nThis is useful, as there were several months in 2011 when LBCCHIP3 was inoperable.\n\n## Some things that might go wrong:\n\nTesting has revealed some occasional issues with the astrometric solution for the individual chips. This can be difficult to diagnose. The registration step using `SWarp` can warn you of some obvious cases, and these can subsequently be removed before rerunning the `SWarp` step by doing, e.g.:\n```\nipython> go_register(fltr_dirs, do_sextractor=False,\n            do_scamp=False, do_SWarp=True)\n```\n\nThere are several issues related to missing or inappropriate files that the current code does not deal with gracefully. The most common is missing flat fields or missing configuration files (found in `LBCgo/LBCgo/conf/`).\n\n\n## Credit:\n\nThis pipeline is built on code initially developed by David Sands, and eventually incorporated into scripts made available by Ben Weiner\n(https://github.com/bjweiner/LBC-reduction).\n\n`LBCgo` was designed to simplify the process of LBC reduction, removing the need for IDL or IRAF in favor of Python. This package continues to require `SCAMP`, `SWarp`, and `SExtractor` provided by Emmanuel Bertin (http://astromatic.iap.fr). It makes extensive use of the `astropy`-affiliated package `CCDProc`.\n\n\n## Known bugs / limitations:\n\n* As of yet no tests are performed to separate LBCB / LBCR images taken with the V-BESSEL filter (which exists in both imagers). Care must be taken to avoid having both in the same directory.\n\n* If flat field images are present, but no image is taken in that flat, an unfortunate behavior results (existing flat fields are divided by the unmatched flats).\n\n* Flat field images taken as "test" images, including only a partial read-out of a single CCD, will cause the code to bail without a helpful error message.\n',
    'author': 'Chris Howk',
    'author_email': 'jhowk@nd.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jchowk/LBCgo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#! /usr/bin/env python
#
# modsProc - bias-correct and flat field MODS raw CCD images.
#
# Usage: modsProc [options] rawFile.fits [rawFile2.fits ... rawFileN.fits] flatField.fits
#
# Where:
#   rawFile*.fits  raw MODS 8x3K FITS image(s) to process
#   pixFlat.fits   is a normalized pixel flat to apply to each rawFile.fits
# Output file will be named rawFile_otf.fits
#
# If N images are listed on the command line, (N-1) are raw images and N is
# the flat field.  Beware!
#
# Options:
#   -b       fix bad pixels (see -l)
#   -l file  bad pixel list to use (requires -b)
#   -w nPix  width of the averaging region either side of the
#             bad pixel region (requires -b).  Default: 5 pixels
#   -f       force overwrite of an existing output file ("clobber")
#   --noflip do not flip a Red channel image X-axis (ignored if blue)
#   -v       print verbose debugging info during execution
#   -V       print version info and exit
#
# Description:
#
#   Processes a raw MODS image as follows:
#
#    1. Remove even/odd overscan and trim (see modsBias.py)
#    2. Divide by the flat field.
#    3. If -b is given, fix known bad columns
#    4. If a Red channel image, flip about X to give the same
#       blue-to-red/left-to-right orientation as the blue channel.
#       can be suppressed with --noflip
#
#   Creates rawFile_otf.fits, otf = overscan, trimmed, and flattened
#
# Quadrant Layout of the MODS Science CCDs:
#
#      ++--------+--------++(NAXIS1,NAXSI2)
#      ||   Q3   |   Q4   ||
#      ++--------+--------++
#      ||   Q1   |   Q2   ||
# (1,1)++--------+--------++
#
#   Each MODS quadrant has even 'e' and odd 'o' readout chains,
#   for a total of 8 readout channels.
#
#   NOTE: This version hardwires the extraction of overscan (really
#   "prescan") from full-frame 8x3K images.  A future revision will
#   use # header info on the overscan columns, once the CCD control
#   software # permits use of overscan (this works using the relevant
#   FITS header # cards, but is currently disabled by a stubborn
#   software bug in the # CCD readout program that is messing up the
#   overscan delacing on # readout).
#
# Authors:
#   Paul Martini (original for R4K CCD at MDM)
#   Rick Pogge (modification for MODS 8x3K CCDs)
#
# Module Dependencies:
#   numpy
#   pyfits
#
# Modification History
#  2011 Sep 08: initial version for just bias subtraction
#  2011 Sep 12: tested, adapted to run on MDM computers, work on MDM4K data
#  2011 Oct 04: adapted for MODS from original proc4k.py by Paul Martini,
#               with significant changes for MODS
#  2011 Oct 21: beta-release version [rwp/osu]
#  2012 Jan 24: minor modifications [rwp/osu]
#  2012 Mar 20: added getopt command-line parsing and various updates [rwp/osu]
#  2012 Apr 01: added bad pixel fixing and other options [rwp/osu]
#  2013 Dec 01: allow multiple raw images (version 2 start) [rwp/osu]
#  2014 Aug 24: using astropy.io.fits instead of unsupported pyfits and
#               suppressing nuisance FITS header warnings [rwp/osu]
# -----------------------------------------------------------------------------

import string as str
import os
from sys import argv, exit
import getopt
import numpy as np
# import pyfits as fits
from astropy.io import fits

# Version number and date

versNum = "2.0.4"
versDate = "2016-03-08"

# Global Defaults

biasSingle = 0
biasRow = 1
biasFit = 2
biasType = biasSingle
outSuffix = '_otf'

avgSize = 5  # Default size of the averaging region for bad columns
modsDir = os.getenv("MODS_DBDIR", "./")  # Default database directory

# Runtime flags

Verbose = False  # Verbose debugging output off
NoClobber = True  # default: no clobber on output
fixBadPix = False  # default: do not fix bad pixels
userBPL = False  # default: use the default bad pixel list
redFlip = True  # default: flip red images about X

# Suppress nuisance warning from astropy.io.fits, which can
# be unusually strict about FITS standard, but the gunk it
# complains about is otherwise harmless (blank keywords)

import warnings

warnings.filterwarnings('ignore', category=UserWarning, append=True)
warnings.filterwarnings('ignore', category=RuntimeWarning, append=True)
warnings.filterwarnings('ignore', category=FutureWarning, append=True)
from astropy.utils.exceptions import AstropyWarning

warnings.filterwarnings('ignore', category=AstropyWarning, append=True)


# ----------------------------------------------------------------
#
# printUsage - print a simple usage message
#

def printUsage():
    print "\nUsage: modsProc [options] rawFile(s).fits flatField.fits"
    print "\nWhere:"
    print "  rawFile(s).fits  raw MODS 8x3K FITS image(s)"
    print "  pixFlat.fits     normalized pixel flat to apply to all rawFiles"
    print "Output file will be named rawFile%s.fits" % (outSuffix)
    print "For N files listed on the command line:"
    print "  1..N-1 are the raw image files"
    print "  N is the flat field for *all* raw images"
    print "\nOptions:"
    print "  -b       fix bad pixels (see -l)"
    print "  -l file  bad pixel list to use (requires -b)"
    print "  -w nPix  width of the averaging region either side of the"
    print "            bad pixel region (requires -b).  Default: %d pixels" % (avgSize)
    print "  -f       force overwrite of an existing output file (\"clobber\")"
    print "  --noflip do not flip a Red channel image X-axis (ignored if blue)"
    print "  -v       print verbose debugging info during execution"
    print "  -V       print version info and exit\n"
    print "\nEnvironment:"
    print "  MODS_DBDIR = directory path to MODS database files"
    print "               currently MODS_DBDIR=%s\n" % (modsDir)


# ----------------------------------------------------------------
#
# readBadPixList(file) - read contents of a bad pixel list
#
# IRAF-format bad pixel coordinate list:
#    xbegin xend ybegin yend
# comments are delimited with #
#

def readBadPixList(file):
    F = open(file, "r")
    M = F.readlines()
    F.close()

    xs = []
    xe = []
    ys = []
    ye = []

    for i in range(len(M)):
        if str.find(M[i], "#") < 0:
            xs.append(float(M[i].split()[0]))
            xe.append(float(M[i].split()[1]))
            ys.append(float(M[i].split()[2]))
            ye.append(float(M[i].split()[3]))

    if len(xs) == 0:
        print "\n** ERROR: %s has no bad regions listed" % (file)
        print "          modsProc aborting.\n"
        exit(1)

    return xs, xe, ys, ye


# ----------------------------------------------------------------
#
# Main Program starts here...
#

# Parse the command-line arguments (GNU-style getopt)

try:
    opts, files = getopt.gnu_getopt(argv[1:], 'fvVbl:w:', ['clobber', 'verbose', 'version',
                                                           'badpix', 'bpl', 'list', 'width',
                                                           'noflip', ])

except getopt.GetoptError, err:
    print "\n** ERROR: %s" % (err)
    printUsage()
    exit(2)

if len(opts) == 0 and len(files) == 0:
    printUsage()
    exit(1)

# Quick Defaults as needed

numAvg = avgSize

for opt, arg in opts:
    if opt in ('-l', '--bpl', '--list'):
        bplFile = arg
        userBPL = True
    elif opt in ('-b', '--badpix'):
        fixBadPix = True
    elif opt in ('-w', '--width'):
        numAvg = int(arg)
        if numAvg < 3:
            print "** ERROR: width of the averaging region must be 3 pixel or larger"
            exit(1)
    elif opt in ('-f', '--clobber'):
        NoClobber = False
    elif opt in ('--noflip'):
        redFlip = False
    elif opt in ('-v', '--verbose'):
        Verbose = True
    elif opt in ('-V', '--version'):
        print "modsBias.py v%s [%s]" % (versNum, versDate)
        exit(0)

if len(files) < 2:
    printUsage()
    exit(0)

# Number of files on the command line

numFiles = len(files)
numRaw = numFiles - 1

# The flat file is the last in the line

flatFile = files[numFiles - 1]

if not os.path.isfile(flatFile):
    print "Cannot find flat field FITS image %s" % (flatFile)
    printUsage()
    exit(0)

# Open and load the flat field image

flatData = fits.open(flatFile, uint=False)

# Process all the raw files on the command line

for i in range(numRaw):
    rawFile = files[i]

    # Create the output filename and test for clobber/no-clobber

    outFile = rawFile[:str.find(rawFile, '.fits')] + outSuffix + '.fits'
    if os.path.isfile(outFile):
        if NoClobber:
            print "\n** ERROR: Operation would overwrite existing FITS file %s" % (outFile)
            print "          Use -f to force output (\"clobber\")."
            print "          modsPixFlat aborting.\n"
            exit(1)
        else:
            if Verbose:
                print "** WARNING: Overwriting existing FITS file %s" % (outFile)
            os.remove(outFile)

    print "Processing image %d of %d: %s --> %s" % (i + 1, numRaw, rawFile, outFile)

    # Open the raw FITS file and extract useful info from the FITS header

    rawData = fits.open(rawFile, uint=False)
    naxis1 = rawData[0].header['NAXIS1']
    naxis2 = rawData[0].header['NAXIS2']
    channel = rawData[0].header['CHANNEL']
    overscanx = rawData[0].header['OVERSCNX']
    overscany = rawData[0].header['OVERSCNY']  # should be 0
    detector = rawData[0].header['DETECTOR']
    telescope = rawData[0].header['TELESCOP']
    instID = rawData[0].header['INSTRUME']

    # If fixing bad pixels (-b), open and read the bad pixel list

    if fixBadPix:
        if userBPL:
            pathBits = os.path.split(bplFile)
            if len(pathBits[0]) == 0:
                bplFile = os.path.join(modsDir, bplFile)
            else:
                # bplFile = os.path.join(modsDir,str.lower(instID)+".bpl") ##CBW
                bplFile = bplFile  ##CBW
        else:
            bplFile = os.path.join(modsDir, str.lower(instID) + ".bpl")

        if os.path.isfile(bplFile) == 0:
            print "\n** ERROR: MODS bad pixel list file %s not found." % (bplFile)
            print "          modsProc aborting.\n";
            exit(1)

        # read and parse the contents of bplFile

        xs, xe, ys, ye = readBadPixList(bplFile)

        if Verbose:
            print "Bad pixel list %s contains %d bad pixel regions to fix" % (bplFile, len(xs) + 1)

    # Stupid bit of hardwiring for now, until the DOS IC code can read
    # X and Y overscan pixels for any subframe ROI without crashing
    # and burning... [rwp]

    if naxis1 == 8288:
        overscanx = 48
        overscany = 0
    else:
        print "** ERROR: NAXIS1=%d, modsProc only implemented for full-frame readout" % (naxis1)
        print "          at the present time, sorry."
        exit(1)

    if Verbose:
        print "Processing %s[%d:%d] OVERSCANX=%d OVERSCANY=%d" % (rawFile, naxis1, naxis2, overscanx, overscany)
        print "Channel: %s" % (channel)
        print "Detector: %s" % (detector)
        print "Telescope: %s" % (telescope)

    if overscanx < 32:
        print "** ERROR: NAXIS1=%d OVERSCNX=%d less than 32 in %s" % (naxis1, overscanx, rawFile)
        print "          modsProc only works with full-frame images for now"
        exit(1)
    if overscany > 0:
        print "Error: code not tested with OVERSCNY > 0!"
        exit(1)

    # Channel-specific setup, if needed, would go here...

    # CCD image "metric" parameters

    c1 = overscanx  # first image column counting from *zero*
    c2 = int(0.5 * naxis1) - 1  # last image column on first half
    c3 = c2 + 1  # first image column on second half
    c4 = naxis1 - overscanx - 1  # last image column
    r1 = overscany  # first image row
    r2 = int(0.5 * naxis2) - 1  # last image row on first half
    r3 = r2 + 1  # first image row on second half
    r4 = naxis2 - overscany - 1  # last image row
    outnaxis1 = c4 - c1 + 1  # columns in output, trimmed image
    outnaxis2 = r4 - r1 + 1  # rows in output, trimmed image
    collen = int(0.5 * outnaxis1)  # number of rows in an image quadrant
    rowlen = int(0.5 * outnaxis2)  # number of rows in an image quadrant

    if Verbose:
        print "Quadrants:"
        print "  Q1: [%d:%d,%d:%d]" % (c1, c2, r1, r2)
        print "  Q2: [%d:%d,%d:%d]" % (c3, c4, r1, r2)
        print "  Q3: [%d:%d,%d:%d]" % (c1, c2, r3, r4)
        print "  Q4: [%d:%d,%d:%d]" % (c3, c4, r3, r4)

        # Calculate the bias level for each amplifier

    data = rawData[0].data

    # Identify the columns to use to calculate the bias level
    # Skip the first and last columns of the overscan

    # Identify the even and odd columns in the overscan

    cols_over_q1e = np.arange(4, overscanx - 4, 2)
    cols_over_q1o = np.arange(5, overscanx - 4, 2)
    cols_over_q2e = np.arange(naxis1 - overscanx + 4, naxis1 - 4, 2)
    cols_over_q2o = np.arange(naxis1 - overscanx + 5, naxis1 - 4, 2)
    cols_over_q3e = cols_over_q1e
    cols_over_q3o = cols_over_q1o
    cols_over_q4e = cols_over_q2e
    cols_over_q4o = cols_over_q2o

    # Identify the even and odd columns in each quadrant

    cols_q1e = np.arange(c1, c2, 2)
    cols_q1o = np.arange(c1 + 1, c2 + 2, 2)
    cols_q2e = np.arange(c3, c4, 2)
    cols_q2o = np.arange(c3 + 1, c4 + 2, 2)
    cols_q3e = cols_q1e
    cols_q3o = cols_q1o
    cols_q4e = cols_q2e
    cols_q4o = cols_q2o

    # Create arrays with the median overscan vs. row for each amplifier

    bias_q1e = np.zeros(rowlen, dtype=float)
    bias_q1o = np.zeros(rowlen, dtype=float)
    bias_q2e = np.zeros(rowlen, dtype=float)
    bias_q2o = np.zeros(rowlen, dtype=float)
    bias_q3e = np.zeros(rowlen, dtype=float)
    bias_q3o = np.zeros(rowlen, dtype=float)
    bias_q4e = np.zeros(rowlen, dtype=float)
    bias_q4o = np.zeros(rowlen, dtype=float)

    # Calculate 1-D bias arrays for each amplifier

    for i in range(r1, r2 + 1, 1):
        bias_q1e[i] = np.median(data[i, cols_over_q1e])  # data[rows, columns]
        bias_q1o[i] = np.median(data[i, cols_over_q1o])
        bias_q2e[i] = np.median(data[i, cols_over_q2e])
        bias_q2o[i] = np.median(data[i, cols_over_q2o])
        bias_q3e[i] = np.median(data[i + rowlen, cols_over_q3e])
        bias_q3o[i] = np.median(data[i + rowlen, cols_over_q3o])
        bias_q4e[i] = np.median(data[i + rowlen, cols_over_q4e])
        bias_q4o[i] = np.median(data[i + rowlen, cols_over_q4o])

    # Subtract the bias from the output

    OverscanKeyValue = 'biasSingle'

    # Subtract a single bias value for each amplifier

    bq1e = np.median(bias_q1e)
    bq1o = np.median(bias_q1o)
    bq2e = np.median(bias_q2e)
    bq2o = np.median(bias_q2o)
    bq3e = np.median(bias_q3e)
    bq3o = np.median(bias_q3o)
    bq4e = np.median(bias_q4e)
    bq4o = np.median(bias_q4o)

    data[r1:r2 + 1, cols_q1e] -= bq1e
    data[r1:r2 + 1, cols_q1o] -= bq1o
    data[r1:r2 + 1, cols_q2e] -= bq2e
    data[r1:r2 + 1, cols_q2o] -= bq2o
    data[r3:r4 + 1, cols_q3e] -= bq3e
    data[r3:r4 + 1, cols_q3o] -= bq3o
    data[r3:r4 + 1, cols_q4e] -= bq4e
    data[r3:r4 + 1, cols_q4o] -= bq4o

    # Write the output file to the FITS array

    rawData[0].data = data[r1:r4 + 1, c1:c4 + 1]
    OverscanKeyComment = 'Overscan subtracted by modsProc.py'

    rawData[0].header.add_history('modsProc v%s %s' % (versNum, versDate))
    rawData[0].header['BIASPROC'] = (OverscanKeyValue, OverscanKeyComment)
    rawData[0].header['BIASQ1E'] = (bq1e, 'Bias subtracted from Q1E')
    rawData[0].header['BIASQ1O'] = (bq1o, 'Bias subtracted from Q1O')
    rawData[0].header['BIASQ2E'] = (bq2e, 'Bias subtracted from Q2E')
    rawData[0].header['BIASQ2O'] = (bq2o, 'Bias subtracted from Q2O')
    rawData[0].header['BIASQ3E'] = (bq3e, 'Bias subtracted from Q3E')
    rawData[0].header['BIASQ3O'] = (bq3o, 'Bias subtracted from Q3O')
    rawData[0].header['BIASQ4E'] = (bq4e, 'Bias subtracted from Q4E')
    rawData[0].header['BIASQ4O'] = (bq4o, 'Bias subtracted from Q4O')

    # We need a working copy of the data from here out

    workData = rawData[0].data

    # Divide the bias-corrected image by the flat field

    if Verbose:
        print "Dividing bias-corrected %s by flat field %s..." % (rawFile, flatFile)

    workData /= flatData[0].data
    rawData[0].header.add_history('modsProc - divided by flat field')
    rawData[0].header['FLATPROC'] = (flatFile, 'Flat field image')

    if fixBadPix:
        # Page through the bad pixel list and fix the pixel
        rawData[0].header.add_history('modsProc - fixing bad pixels')
        rawData[0].header['BPLFILE'] = (bplFile)
        rawData[0].header['BPWIDTH'] = (numAvg, 'Bad pixel averaging region width [pixels]')

        for i in range(len(xs)):
            xstart = int(xs[i] - 1)
            xend = int(xe[i])
            ystart = int(ys[i] - 1)
            yend = int(ye[i])
            nx = xe[i] - xs[i] + 1

            if Verbose:
                print "  Fixing pixels in region [%d:%d,%d:%d]" % (xs[i], xe[i], ys[i], ye[i])

            # The "fix" is to replace bad-column pixels by the median of
            # pixels +/-numAvg either side of the bad column(s).

            xsl = xstart - numAvg
            xel = xstart - 1
            xsr = xend + 1
            xer = xend + numAvg

            for y in range(ystart, yend, 1):
                medLeft = np.median(workData[y, xsl:xel])
                medRight = np.median(workData[y, xsr:xer])
                corrVal = (medLeft + medRight) / 2.0
                if (nx > 1):
                    workData[y, xstart:xend] = corrVal
                else:
                    workData[y, xstart] = corrVal

    # If this is a red-channel image, we need to flip the red channel into
    # blue-to-red wavelength order

    if str.upper(channel) == 'RED':
        if redFlip:
            if Verbose:
                print "Flipping Red-channel CCD into blue-to-red wavelength order..."
            workData = np.fliplr(workData)
            rawData[0].header.add_history('modsProc - flipped image about the X-axis')
        else:
            if Verbose:
                print "Red-channel CCD images were not flipped in X (--noflip)..."

    # Write the processed image data to disk

    rawData[0].data = workData

    if Verbose:
        print "Writing bias- and flat-corrected %s --> %s" % (rawFile, outFile)

    rawData.writeto(outFile, output_verify='ignore')

    # done close the fits file

    rawData.close()

# Bottom of the loop, all done

if Verbose:
    print "modsProc Done"

exit(0)
